"""
This module, has the function writeConfig, to generate a cloudconf.py file
If invoked at the command-line, the default config will be written

Note: You need to manually wipe cloudconf.py before using this!

Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2010 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see
http://www.gnu.org/licenses/lgpl-2.1.html
"""

import sys

def writeConfig(withHidden = False):
    """Direct configmanager to write config
    withHidden controls if hidden variables should be written"""

    cc.genHidden = withHidden

    from scicloud.util.configmanager import ConfigSettings
    config = ConfigSettings(cc.config,do_reload=True)
    cc.flush_config()


if __name__ == '__main__':
    withHidden = False
    if len(sys.argv) > 1 and sys.argv[1] == 'advanced':
        withHidden = True

    writeConfig(withHidden)



"""
Various utility functions

Copyright (c) 2009 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

import sys
import types
import cPickle
import inspect
import datetime
import os

from functools import partial
from warnings import warn

def islambda(func):
    return getattr(func,'func_name') == '<lambda>'

def funcname(func):
    """Return name of a callable (function, class, partial, etc.)"""
    module = ""
    if hasattr(func,'__module__'):
        module = (func.__module__ if func.__module__ else '__main__')
    """Return a human readable name associated with a function"""
    if inspect.ismethod(func):
        nme = '.'.join([module,func.im_class.__name__,func.__name__]) 
    elif inspect.isfunction(func):
        nme =  '.'.join([module,func.__name__])
    elif inspect.isbuiltin(func):
        return  '.'.join([module,func.__name__])
    elif isinstance(func,partial):
        return 'partial_of_' + funcname(func.func)    
    elif inspect.isclass(func):
        nme = '.'.join([module,func.__name__])
        if hasattr(func, '__init__') and inspect.ismethod(func.__init__):            
            func = func.__init__
        else:
            return nme #can't extract more info for classes
        
    else:
        nme = 'type %s' % type(func)
        if hasattr(func, '__name__'):
            nme = '%s of %s' % (func.__name__, type(func))
        return nme
    nme +=  ' at ' + ':'.join([func.func_code.co_filename,str(func.func_code.co_firstlineno)])
    return nme 
    
def min_args(func):
    """Return minimum (required) number args this function has"""
    if inspect.isfunction(func):
        op_args = len(func.func_defaults) if func.func_defaults else 0 
        return func.func_code.co_argcount - op_args
    elif inspect.ismethod(func):
        return min_args(func.im_func) - 1
    elif inspect.isclass(func):
        if hasattr(func, '__init__'): #check class constructor
            return min_args(func.__init__)
        else:
            return 0
            
    raise TypeError('cannot deal with type: %s' % type(func))

def max_args(func):
    """Return maximum (required + default) number of arguments callable can take"""
    if inspect.isfunction(func):
        return func.func_code.co_argcount
    elif inspect.ismethod(func):
        return max_args(func.im_func) - 1
    elif inspect.isclass(func) and hasattr(func, '__init__'): #check class constructor
        if hasattr(func, '__init__'): #check class constructor
            return max_args(func.__init__)
        else:
            return 0
    raise TypeError('cannot deal with type: %s' % type(func))


def getargspec(func):
    """Returns an argspec or None if it can't be resolved
    Our argspec is similar to inspect's except the name & if it is a method is appended as the first argument
    Returns (name, is_method, args, *args, **kwargs, defaults)
    """
    
    try:
        argspec = inspect.getargspec(func)
    except TypeError:
        return None

    out_list = [func.__name__, int(inspect.ismethod(func))]
    out_list.extend(argspec)
    return out_list
    
def validate_func_arguments(func, test_args, test_kwargs):
    """First pass validation to see if args/kwargs are compatible with the argspec
    Probably doesn't catch everything that will error
    Known to miss:
        Validate that anonymous tuple params receive tuples
    
    This is only valid for python 2.x
    
    Returns true if validation passed; false if validation not supported
        Exception raised if validation fails
    """
    try:
        argspec = inspect.getargspec(func)
    except TypeError: #we can't check non-functions
        return False
        
    return validate_func_arguments_from_spec( (func.__name__, int(inspect.ismethod(func))) + argspec, 
                                              test_args, 
                                              test_kwargs.keys())
    
    
def validate_func_arguments_from_spec(argspec, test_args, test_kwargs_keys):
    
    name, is_method, args, varargs, varkw, defaults = argspec
    
    if defaults == None:
        defaults = []
    else:
        defaults = list(defaults)
    
    if is_method: #ignore self/cls
        args = args[1:]
                
    
    name += '()' #conform to python error reporting
    test_args_len = len(test_args)   
    
    #kwd exist?
    if not varkw:
        for kw in test_kwargs_keys:
            if kw not in args:
                raise TypeError("%s got an unexpected keyword argument '%s'" % (name, kw))
        
    #kwd not already bound by passed arg?
    kwd_bound = args[test_args_len:] #These must all be default or bound to kwds
    if not varkw:
        for kw in test_kwargs_keys:
            if kw not in kwd_bound:
                raise TypeError("%s got multiple values for keyword argument '%s'" % (name, kw))
    
    #verify argument count
    firstdefault = len(args) - len(defaults)
    nondefargs = args[:firstdefault]
    
    defaults_injected = 0
    for kw in test_kwargs_keys:
        if kw in nondefargs:
            defaults.append(None)  #pretend another default is there for counting
            defaults_injected += 1

    min = len(args) - len(defaults)
    max = len(args)
    
    #correct for default injection
    min+=defaults_injected
    max+=defaults_injected
    test_args_len += defaults_injected
    
    if varargs:
        max = sys.maxint
    if min < 0:
        min = 0
    
    if test_args_len < min or max < test_args_len:
        err_msg = '%s takes %s arguments (%d given)'
        
        if min == max:
            arg_c_msg = 'exactly %s' % min        
        elif test_args_len < min:
            arg_c_msg = 'at least %s' % min
        else:
            arg_c_msg = 'at most %s' % max
        
        raise TypeError(err_msg % (name, arg_c_msg, test_args_len))
    
    return True

def fix_time_element(dct, key):
    """Fix time elements in dictionaries coming off the wire"""
    item = dct.get(key)
    if item == 'None': #returned by web instead of a NoneType None
        item = None
        dct[key] = item
    if item:
        dct[key] = datetime.datetime.strptime(item,'%Y-%m-%d %H:%M:%S')
    return dct

def fix_sudo_path(path):
    """Correct permissions on path if using sudo from another user and keeping old users home directory"""
    
    if os.name != 'posix': 
        return
    
    sudo_uid = os.environ.get('SUDO_UID')
    sudo_user = os.environ.get('SUDO_USER')
    
    if sudo_uid != None and sudo_user:
        sudo_uid = int(sudo_uid)
        home = os.environ.get('HOME')
        sudo_user_home = os.path.expanduser('~' + sudo_user)
        
        # important: Only make modifications if user's home was not changed with sudo (e.g. sudo -H)
        if home == sudo_user_home:
            sudo_gid = os.environ.get('SUDO_GID')
            sudo_gid = int(sudo_gid) if sudo_gid else -1
            try:
                os.chown(path, sudo_uid, sudo_gid)
            except Exception, e:
                warn('PiCloud cannot fix SUDO Paths. Error is %s:%s' % (type(e), str(e)))
                
   
"""Ordered Dictionary"""
import UserDict
class OrderedDict(UserDict.DictMixin):
    
    def __init__(self, it = None):
        self._keys = []
        self._data = {}
        if it:
            for k,v in it:
                self.__setitem__(k,v)
        
        
    def __setitem__(self, key, value):
        if key not in self._data:
            self._keys.append(key)
        self._data[key] = value
    
    def insertAt(self, loc, key, value):
        if key in self._data:
            del self._data[self._data.index(key)]
        self._keys.insert(loc, key)
        self._data[key] = value
        
    def __getitem__(self, key):
        return self._data[key]
    
    
    def __delitem__(self, key):
        del self._data[key]
        self._keys.remove(key)
        
        
    def keys(self):
        return list(self._keys)    
    
    def copy(self):
        copyDict = OrderedDict()
        copyDict._data = self._data.copy()
        copyDict._keys = self._keys[:]
        return copyDict
    
"""Python 2.5 support"""
from itertools import izip, chain, repeat
if sys.version_info[:2] < (2,6):
    def izip_longest(*args):
        def sentinel(counter = ([None]*(len(args)-1)).pop):
            yield counter()         # yields the fillvalue, or raises IndexError
        fillers = repeat(None)
        iters = [chain(it, sentinel(), fillers) for it in args]
        try:
            for tup in izip(*iters):
                yield tup
        except IndexError:
            pass

if __name__ == '__main__':
    """Validate the validate_func_arguments function"""
    def foo0():
        pass
    
    def foo1(a):
        pass
    
    def foo2(a, b=2):
        pass

    def foo21(a, b):
        pass
    
    def foo3(a, (x,y), b):
        """lovely anonymous function"""
        pass
    
    def consist(func, *args, **kwargs):
        typerror = None
        try:
            func(*args, **kwargs)
        except TypeError, e:
            typerror = e
        
        print '%s %s %s' % (func, args, kwargs)
        try:
            validate_func_arguments(func, args, kwargs)
        except TypeError, e:
            if not typerror:
                print 'unexpected typerror! %s' % str(e)
                raise
            else:
                print '%s == %s' % (typerror, str(e))
        else:
            if typerror:
                print 'missed error! %s' % typerror
                raise
            else:
                print 'no error!'
                
    consist(foo0)
    consist(foo0, 2)
    consist(foo0, k=2)
    consist(foo0, 3, k=4)
    
    consist(foo1)
    consist(foo1, b=2)
    consist(foo1, a=2)
    consist(foo1, 2)
    consist(foo1, 3)
    consist(foo1, 3, a=2)
    consist(foo1, 3, b=2)
    
    consist(foo2)
    consist(foo2, b=2)
    consist(foo2, b=2, c=3)
    consist(foo2, a=2)
    consist(foo2, a=2, b=2)
    consist(foo2, a=2, b=2, c=3)
    consist(foo2, 2, a=10)
    consist(foo2, 3)
    consist(foo2, 3, 4)
    consist(foo2, 3, 4, 7)
    consist(foo2, 3, b=2)
    consist(foo2, 3, a=10, b=2)
    consist(foo2, 3, b=2, c=2)
    consist(foo2, 3, a=10, b=2, c=4)
    
    consist(foo21, 3, 4)
    consist(foo21, 3, b=4)
    consist(foo21, a=3, b=4)
    consist(foo21, b=4)
    consist(foo21, a=4)
    consist(foo21)
    consist(foo21, 4, 3, 5)
    
    consist(foo3, 2, (4,3), 9)
    consist(foo3, 2, (4,3), b=9)
    consist(foo3, 2, (4,3), a=9)
    consist(foo3, 2, (4,3), a=9, b=9)
    consist(foo3, 2,  a=9, b=9)
    consist(foo3, 2, (4,3))
    
    #we can't catch below..
    #consist(foo3, 2,  10, 12)
    



from __future__ import with_statement

'''
Provides for storage and retrieval of PiCloud credentials

Current credentials include:
- cloudauth: key/secretkey
- ssh private keys (environments/volumes) 
'''
import distutils
import os

from ConfigParser import RawConfigParser
from .. import cloudconfig as cc
import logging
cloudLog = logging.getLogger('Cloud.credentials')


credentials_dir = os.path.expanduser(os.path.join(cc.baselocation,'credentials'))

"""general"""
key_cache = {}  # dictionary mapping key to all authentication information

def save_keydef(key_def, api_key=None):
    """Save key definition to necessary files. Overwrite existing credential
        If *api_key* not None, verify it matches key_def
    """
    key_def['api_key'] = int(key_def['api_key'])
    if not api_key: 
        api_key = key_def['api_key']
    else:    
        assert (key_def['api_key'] == int(api_key))    
    key_cache[api_key] = key_def
    write_cloudauth(key_def) #flush authorization
    write_sshkey(key_def)    #flush ssh key
    
def download_key_by_key(api_key, api_secretkey):
    """Download and cache key""" 
    api_key = int(api_key)       
    from ..account import get_key_by_key
    key_def = get_key_by_key(api_key, api_secretkey)
    cloudLog.debug('Saving key for api_key %s' % api_key)
    save_keydef(key_def, api_key)
    return key_def

def download_key_by_login(api_key, username, password):
    """Download and cache key by using PiCloud login information""" 
    api_key = int(api_key)       
    from ..account import get_key
    key_def = get_key(username, password, api_key)
    save_keydef(key_def, api_key)
    return key_def

def verify_key(api_key):
    """Return true if we have valid sshkey and cloudauth for this key.
    False if any information is missing"""
    key_def = key_cache.get(api_key, {})
    if 'api_secretkey' not in key_def:    
        if not resolve_secretkey(api_key):
            cloudLog.debug('verify_key failed: could not find secretkey for %s', api_key)
            return False
    if not 'private_key' in key_def:
        res = verify_sshkey(api_key)
        if not res:
            cloudLog.debug('verify_key failed: could not find sshkey for %s', api_key)
        return res

def get_credentials_path(api_key):
    """Resolve directory where credentials are stored for a given api_key
    Create directory if it does not exist"""
    path = os.path.join(credentials_dir, str(api_key))
    try:
        distutils.dir_util.mkpath(path)
    except distutils.errors.DistutilsFileError:
        cloudLog.exception('Could not generate credentials path %s' % path)
    return path

""" Api keys"""
#constants:
api_key_section = 'ApiKey'

def get_cloudauth_path(api_key):
    """Locate cloudauth path"""
    base_path = get_credentials_path(api_key)
    return os.path.join(base_path, 'cloudauth')

def read_cloudauth(api_key):
    """Load cloudauth for api_key"""
    path = get_cloudauth_path(api_key)
    if not os.path.exists(path):
        raise IOError('path %s not found' % path)
    config = RawConfigParser()
    config.read(path)
    
    key_def = key_cache.get(api_key, {})
    key = config.getint(api_key_section, 'key')
    if key != api_key:
        raise ValueError('Cloudauth Credentials do not match. Expected key %s, found key %s' % (api_key, key))
    key_def['api_key'] = key
    key_def['api_secretkey'] = config.get(api_key_section, 'secretkey')
    key_cache[int(api_key)] = key_def
    return key_def

def get_saved_secretkey(api_key):
    """Resolve the secret key for this api_key from the saved cloudauth credentials"""
    api_key = int(api_key)
    key_def = key_cache.get(api_key)
    if not key_def:
        key_def = read_cloudauth(api_key)
    return key_def['api_secretkey']

def write_cloudauth(key_def):
    """Write key/secret key information defined by key_def into cloudauth"""
    api_key = str(key_def['api_key'])
    api_secretkey = key_def['api_secretkey']    
    path = get_cloudauth_path(api_key)
        
    config = RawConfigParser()
    config.add_section(api_key_section)
    config.set(api_key_section, 'key', api_key)
    config.set(api_key_section, 'secretkey', api_secretkey)
    try:
        with open(path, 'wb') as configfile:
            config.write(configfile)
    except IOError, e:
        cloudLog.exception('Could not save cloudauth credentials to %s' % path)
    try:
        os.chmod(path, 0600)
    except:
        cloudLog.exception('Could not set permissions on %s' % path)
        

def resolve_secretkey(api_key):
    """Find secretkey for this api_key
    Return None if key cannot be found
    """    
    try:
        secretkey = get_saved_secretkey(api_key)
    except Exception, e:
        if not isinstance(e, IOError):
            cloudLog.exception('Unexpected error reading credentials for api_key %s' % api_key)
        return None
    else:
        return secretkey        


""" SSH private keys 
These private keys are used to connect to PiCloud
"""    

def get_sshkey_path(api_key):
    """Locate where SSH key is stored"""
    base_path = get_credentials_path(api_key)
    return os.path.join(base_path,'id_rsa')

def read_sshkey(api_key):
    """Read sshkey from file.
    Save to cache and return key_def. key will be in key_def['private_key']"""
    path = get_sshkey_path(api_key)    
    with open(path, 'rb') as f:
        private_key = f.read()
    key_def = key_cache.get(api_key, {})
    key_def['api_key'] = api_key
    key_def['private_key'] = private_key
    key_cache[int(api_key)] = key_def
    return key_def
    
def verify_sshkey(api_key):
    """Verify sshkey presence
    Todo: Actually validate key    
    """
    path = get_sshkey_path(api_key)
    if os.path.exists(path):
        try:
            os.chmod(path, 0600)
        except:
            cloudLog.exception('Could not set permissions on %s' % path)
        return True
    return False

def write_sshkey(key_def):
    """Save key_def['private_key'] to sshkey_path"""
    private_key = key_def['private_key']
    api_key = key_def['api_key']
    path = get_sshkey_path(api_key)    
    try:
        with open(path, 'wb') as f:
            f.write(private_key)
    except IOError, e:
        cloudLog.exception('Could not save ssh private key to %s' % path)
    else:
        try:
            os.chmod(path, 0600)
        except:
            cloudLog.exception('Could not set permissions on %s' % path)
        
    
def test(key, secretkey):
    key_has = verify_key(key)
    print 'have key already? %s' % key_has 
    if not key_has:
        print 'downloading'
        download_key_by_key(key, secretkey)
        key_has = verify_key(key)
        print 'have key now? %s' % key_has

    secretkey = resolve_secretkey(key)
    print 'your key is %s' % secretkey
        


"""
Serialization submodule
Responsible for all cloud related serialization routines

Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2009 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

from .serializationhandlers import Serializer, CloudSerializer, DebugSerializer, Deserializer
from .report import SerializationReport

try:
    import cPickle as pickle
except ImportError:
    import pickle as pickle

def serialize(obj, needsPyCloudSerializer = False, useDebugSerializer = False):
    """Pickle obj into a string.  
    If needsPyCloudSerializer is set, additional types are allowed at the cost of speed
    If useDebugSerializer is set, exceptions are more detailed, again at the cost of speed
    """
    
    if not needsPyCloudSerializer:
        return Serializer(obj).run_serialization()
    elif not useDebugSerializer:
        return CloudSerializer(obj).run_serialization()
    else:
        return DebugSerializer(obj).run_serialization()
    
def deserialize(str):
    """Deserialize the serialized object described by string str"""
    return Deserializer(str).deserializedObj



"""
Caching objects

Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2010 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""


from __future__ import with_statement 
import threading
import weakref
import collections

from .xrange_helper import maybe_xrange_iter 

#unique object to identify when attribute not present
#By design this is designed to evaluate to False
_NO_ATTRIBUTE = list()

class JobCacheItem(object):
    """Stub cache item.
    Attributes are dynamically added by the managers
    
    USAGE NOTE: None is reserved for a non-existent field.  One cannot cache None
    """
    
    def update(self, adict):
        """Update attributes with adict"""
        for key, val in adict.items():
            setattr(self,key,val)

class JobAbstractCacheManager(object):
    """
    Abstract Cache manager that does not cache
    """
    
    def getCached(self, jids, fields):
        """
        Request cached information about jids        
        return sequence of respective jid->field mapping for each field respectively
        """        
        return  {}
    
    def putCached(self, jid, **kwargs):
        """
        Cache a jid with certain properties
        """           
        pass
    
    def deleteCached(self, jids):
        """
        Remove these jids from the cache
        """
        pass        

class JobCacheManager(JobAbstractCacheManager):
    """
    Simple manager for dealing with cloud cache
    """
    
    jobMap = None #maps jids to cache
    cacheLock = None 
    myItem = JobCacheItem  #type of cache item to generate
    
    def __init__(self):        
        self.jobMap = {} #maps jids to cache
        self.cacheLock = threading.RLock()        
        
    def _getJob(self, jid):
        return self.jobMap.get(jid, None)
    
    def getCached(self, jids, fields):
        """
        Request cached information about jids        
        return dictionary mapping jid to dictionary describing cached fields
        If no cached fields, jid will not appear in dictionary
        """
        outdict = {}
                       
        for jid in maybe_xrange_iter(jids):
            job = self._getJob(jid)
            if job:
                jf = {}
                for field in fields:
                    attr = getattr(job,field,_NO_ATTRIBUTE)
                    if attr != _NO_ATTRIBUTE:                 
                        jf[field] = attr
                if jf:
                    outdict[jid]  = jf    
        return outdict
    
    def _deleteJob(self, jid):
        try:
            del self.jobMap[jid]
        except KeyError:
            pass
        
    
    def deleteCached(self, jids):
        """
        Remove these jids from the cache
        """
        with self.cacheLock:
            for jid in jids:
                self._deleteJob(jid)
    
    def _addJob(self, jid, item):
        self.jobMap[jid] = item
        return item
            
    def putCached(self, jid, **kwargs):    
        """
        Cache a jid with certain properties
        """    
        with self.cacheLock:    
            job = self._getJob(jid)
            if not job:
                job = self.myItem()
                needAdd = True
            else:
                needAdd = False
            job.update(kwargs)
            if needAdd:
                return self._addJob(jid, job)
            return job

class JobFiniteCacheManager(JobCacheManager):
    """
    Regular JobCacheManager which utilizes clock algorithm to limit number of elements in cache
    """
    jobClock = None #clock algorithm to push jobs out
    clockHand = 0
    numDeleted = 0 #indicates number of non-purged jobClock entries

    def __init__(self, cacheSize):
        JobCacheManager.__init__(self)        
        self.cacheSize = cacheSize
        self.jobClock = [] #clock algorithm to push jobs out
        self.clockHand = 0    
        
    def _replaceNotifier(self, oldjid):
        """Notification when oldjid is to be replaced"""
        self.numDeleted-=1
    
    def _replace(self, newjid):
        """Find a spot for newjid"""
        sz = len(self.jobClock)    
        while (True):
            self.clockHand=(1+self.clockHand)%sz
            self.clockHand%=sz
            jid = self.jobClock[self.clockHand]
            if self.numDeleted:  
                #replace deleted entries first
                #technically it is wrong to move the clockHand, but delete is rare
                if jid not in self.jobMap:
                    self.numDeleted -=1
                    self.jobClock[self.clockHand] = newjid                                       
                    return                    
            elif getattr(self.jobMap[jid],'clockBit',0) == 0:
                del self.jobMap[jid]  #don't use _deleteJob as it would increment numDeleted
                self.jobClock[self.clockHand] = newjid
                return
            else:
                self.jobMap[jid].clockBit = 0 
    
    def _getJob(self, jid):
        item = self.jobMap.get(jid, None)
        if item:
            item.clockBit = 1        
        return item
    
    def _addJob(self, jid, item):
        self.jobMap[jid] = item
        if len(self.jobMap) > self.cacheSize:
            self._replace(jid)
        else:
            self.jobClock.append(jid)
        return item

    def _deleteJob(self, jid):
        try:
            del self.jobMap[jid]
        except KeyError:
            pass
        else:
            self.numDeleted+=1


class JobFiniteSizeCacheManager(JobFiniteCacheManager):
    """
    A JobFiniteCacheManager limited by total cache size
    NOTE: This can only store objects that define __len__ (e.g. serialized results)
        __len__ is considered the object's size
        Size of actual cache items is ignored in size calculation
    cacheSize is interpreted as maximum space strings can take
    
    Clock algorithm is still used.
    Items taking >= 0.5*total space are not cached
    
    A double-sided queue is used for the clock     
    """
    
    mySize = 0 #size this cache is consuming
    trackedAttr = "" #Name of attribute of JobCacheItem whose size is being tracked
    
    def __init__(self, cacheSize, trackedAttr):
        JobCacheManager.__init__(self)        
        self.cacheSize = cacheSize
        self.jobClock = collections.deque() 
        self.trackedAttr = trackedAttr
    
    def _makeSpace(self, extraSize):
        """Clear out space"""
        sz = len(self.jobClock)    
        while (self.mySize + extraSize >= self.cacheSize):
            jid = self.jobClock.pop()
            if jid in self.jobMap:      #if deleted, no longer here                    
                if getattr(self.jobMap[jid],'clockBit',0) == 0:
                    self._deleteJob(jid)
                else:
                    self.jobMap[jid].clockBit = 0
                    self.jobClock.appendleft(jid) 
        
    def _deleteJob(self, jid):
        try:
            item = self.jobMap[jid]
            del self.jobMap[jid]
        except KeyError:
            pass
        else:
            self.mySize -= len(getattr(item, self.trackedAttr))
            #jobclock entry is deleted eventually by _makeSpace
    
    def _addJob(self, jid, item):
        
        sz = len(getattr(item, self.trackedAttr))
        if 2*sz >= self.cacheSize: #ignore items that are too large
            return None 
        
        self._makeSpace(sz)
        
        self.jobMap[jid] = item
        self.jobClock.appendleft(jid)
        
        self.mySize += sz
        return item            

class JobFiniteDoubleCacheManager(JobFiniteCacheManager):
    """
    A JobFiniteCacheManager that has weakrefs to items in a JobFiniteSizeCacheManager (childManager)
    weakrefs are used to point to a potentially large data structure that can be ejected by the
        childManager. We still hold the other pieces of data though
    """
    
    weakRefAttr = None  #Item that should be moved to child manager
    childManager = None #child manager that tracks size
    
    _constructedItems = {}  #static dictionary mapping to constructed classes
    
    @classmethod
    def _constructJobItemType(cls, trackedAttr):
        """Construct a JobCacheItem subclass that has weakRefAttr 
        as a property that dereferences the weakref"""
        
        #check cache first
        try:
            return cls._constructedItems[trackedAttr]
        except KeyError:
            pass
        
        hiddenAttr = '_' + trackedAttr #hide weakref
        
        def _generateProp():
            def getProp(self):
                wrefItem = getattr(self, hiddenAttr)
                if not wrefItem:
                    return None                                
                wrefItem = wrefItem() #dereference weakref
                if not wrefItem:
                    return _NO_ATTRIBUTE              
                wrefItem.clockBit = 1 #update clock of item                
                return getattr(wrefItem, trackedAttr) 
            
            #TODO: Setprop (would need access to cache manager)            
            return property(getProp, doc="Access %s through weakref" % trackedAttr)
        
        dct = {trackedAttr: _generateProp()}
        JobItemType = type('JobCacheItem' + trackedAttr, (JobCacheItem,), dct)
        
        cls._constructedItems[trackedAttr] = JobItemType
        return JobItemType
        
        
    
    def __init__(self, cacheSize, childManager):
        JobFiniteCacheManager.__init__(self, cacheSize)
        self.childManager = childManager 
        tracked = childManager.trackedAttr
        self.myItem = self.__class__._constructJobItemType(tracked)
        self.weakRefAttr = tracked  #weakref item is the child's trackedAttr        
        
                        
    def putCached(self, jid, **kwargs):
        #Generate weakref if kwargs includes target        
        if self.weakRefAttr in kwargs:
            newItem = self.childManager.putCached(jid,**{self.weakRefAttr: kwargs[self.weakRefAttr]})
            if newItem: #child is allowed to reject
                #We must make a weakref to the item, as weakrefs to strings are not allowed
                kwargs['_' + self.weakRefAttr] = weakref.ref(newItem)
            del kwargs[self.weakRefAttr] #don't let parent put in cache            
        return JobFiniteCacheManager.putCached(self, jid, **kwargs)
        
    def _deleteJob(self, jid):
        #Delete child's result
        self.childManager.deleteCached([jid])
        JobFiniteCacheManager._deleteJob(self,jid)
        
    def _replaceNotifier(self, oldjid):
        """Notification when oldjid is to be replaced"""
        JobFiniteCacheManager._replaceNotifier(self, oldjid)
        #wipe child entry:
        JobFiniteCacheManager.deleteCached([oldjid])       
        
        
    
    



"""
Interface for cloud - the clientside module for PiCloud

Sample usage::

    import scicloud as cloud
    jid = cloud.call(lambda: 3*3)
    >> Returns a job identifier
    cloud.result(jid)
    >> Returns 9

This will run the function lambda: 3*3 on PiCloud's
cluster, and return the result. Most functions, even
user-defined ones, can be passed through cloud.call

For cloud to work, you must first run 'cloud setup' in your shell.

Alternatively, you can use the simulator by setting use_simulator to True
in cloudconf.py or running cloud.start_simulator()
"""
"""
Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2010 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see
http://www.gnu.org/licenses/lgpl-2.1.html
"""

import sys
import imp
#import threading

#versioning:
from .versioninfo import release_version
__version__ = release_version

import logging

#Input hooks
class ImportHook(object):
    def __init__(self):
        self.mods = []  #Reloading mods in order will reload entire cloud

    def load_module(self, fullname):
        """This is called by the system to load a module
            We use it to track mopdules after their code is loaded"""
        parentmod, ext, submod = fullname.rpartition('.')
        if parentmod:
            path = sys.modules[parentmod].__path__
        else:
            path = ""

        file, filename, tuple = imp.find_module(submod, path)

        #This may load additional modules which are appended before this module in self.mods
        imp.acquire_lock()
        try:
            mod = imp.load_module(fullname, file, filename, tuple)
        except ImportError, i:
            raise
        finally:
            imp.release_lock()

        self.mods.append(fullname)
        return mod


    def find_module(self, fullname, path = None):
        """This is called by the system to find a module.
        cloud.* modules are hooked into our system"""
        #print 'find %s -- %s' % (fullname, threading.current_thread().ident)

        if False:
            v = logging.getLogger('Cloud')
            try:
                v.debug('IMPORT: Attempting to load %s from %s', fullname, path)
            except Exception: # race condition with multiprocessing
                pass

        if fullname.startswith('cloud') and 'server' not in fullname:
            self.path = path
            parentmod, ext, submod = fullname.rpartition('.')
            if parentmod:
                path = sys.modules[parentmod].__path__
            else:
                path = ""
            try: #Fails due to attempting relative import of python packages, e.g cloud.os
                grp = imp.find_module(submod, path)
                if not grp: #import error
                    return None
            except ImportError:
                return None
            else:
                return self
        else:
            return None

_modHook = ImportHook()
sys.meta_path.append(_modHook) #registers our import hook

from . import cloudconfig as cc
import cloudinterface
from . import cloud

__cloud = None
__type = None
__immutable = False

#function bindings
call = None
map = None
status = None
join = None
result = None
kill = None
is_simulated = None
running_on_cloud = None
connection_info = None

__all__ = ["call", "map", "status", "result", "iresult", "is_simulated", "join", "kill",
            "delete", "info", "connection_info", "running_on_cloud", "setkey", "finished_statuses",
             "start_simulator", "config", "getconfigpath", "close","c1","c2","m1"]

c = """The Cloud Simulator simulates cloud.* functions locally, allowing for quicker debugging.
If this is enabled, all cloud.* functions will be run locally, rather than on PiCloud. See web documentation."""
__useSimulator =  cc.account_configurable('use_simulator',
                                          default=False,
                                          comment=c)

def _launch_cloud():
    cloudinterface._setcloud(sys.modules[__name__], 'simulated' if __useSimulator else 'network', restart=True)

_launch_cloud()

from .cloud import CloudException, CloudTimeoutError

def start_simulator(force_restart = False):
    """
    Simulate cloud functionality locally.
    If *force_restart* or not already in simulation mode, restart cloud and enter simulation.
    In simulation mode, the cloud will be run locally, on multiple processors, via
    the multiprocessing library.
    Additional logging information will be enabled.
    For more information, see
    `PiCloud documentation <http://docs.scivm.com/cloud_simulator.html>`_
    """

    cloudinterface._setcloud(sys.modules[__name__], 'simulated', restart=force_restart)


def setkey(api_key, api_secretkey=None, server_url=None, restart=False, immutable=False):
    """
    Connect cloud to your PiCloud account, given your *api_key*.
    Your *api_key* is provided by PiCloud on the
    `API Keys <https://dashboard.scivm.com/accounts/apikeys/>`_ section of the PiCloud website.

    The *api_secretkey* is generally stored on your machine.  However, if you have not previously used
    this api_key or selected it in 'cloud setup', you will need to provide it.

    *server_url* specifies the PiCloud server to connect to.  Leave this blank to auto-resolve servers.

    *restart* forces the cloud to reconnect

    This command will disable the simulator if it is running.
    """

    cloudinterface._setcloud(sys.modules[__name__], 'network', api_key, api_secretkey, server_url, restart, immutable)

def set_dependency_whitelist(whitelist):
    """
    By default all relevant dependencies found will be transported to PiCloud
    In secure scenarios, you may wish to restrict the dependencies that may be transferred.

    whitelist should be a list consisting of module names that can be transported.

    Example::

        ['mypackage','package2.mymodule']

        * Allows mypackage, package2.mymodule
        * Disallows foo_module, package2.mymodule2

    Set to None to re-enable full automatic dependency magic

    .. warning::

        Dependency whitelist will reset if setkey() or start_simulator() is run
    """
    from .transport.adapter import DependencyAdapter
    from .transport.network import HttpConnection

    # must have a connected network connection to create a dependency manager
    netcon = _getcloudnetconnection()
    adapter = netcon.adapter

    adapter._create_dependency_manager(whitelist)


def _getcloud():
    """Return internal cloud object. Only for internal use"""
    return __cloud

def _getcloudnetconnection():
    """Returns cloud connection if it is an HttpConnection. else raise Exception"""
    from .transport.adapter import SerializingAdapter
    from .transport.network import HttpConnection
    if not isinstance(__cloud.adapter, SerializingAdapter):
        raise Exception('Unexpected cloud adapter being used')

    conn = __cloud.adapter.connection
    if isinstance(conn, HttpConnection):
        __cloud._checkOpen()
        return conn
    else:
        raise RuntimeError('Cannot use this functionality in simulation')

def getconfigpath():
    """
    Return the directory where PiCloud configuration and logging files are stored.
    Configuration/logs are stored on a per-user basis
    """
    return _cc.fullconfigpath

"""Read/Write configuration file"""
from .util.configmanager import ConfigSettings
config = ConfigSettings(cc.config)  #sets cc.config values
if cc._needsWrite:
    cc.flush_config()
    cc._needsWrite = False

del ConfigSettings
_cc = cc
del cc

try:
    from . import mp as mp
except ImportError:
    try:
        import multiprocessing
    except ImportError: #if mp error, fall back
        cloud.cloudLog.warn('Multiprocessing is not installed. Cloud.mp will be disabled')
    else:
        raise #something else has gone wrong

from . import account   #cloud.account
from . import cron      #cloud.cron

"""
if __cloud.running_on_cloud():
    # use server-side implementation of cloud.bucket
    from . import _server_bucket as bucket   #cloud.bucket
else:
    from . import bucket    #default loud.bucket
"""

from . import shell     #cloud.shell
from . import queue     #cloud.queue

from . import files     #cloud.files (DEPRECATED)



"""
Defines (de)serialization handlers

Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2011 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

try:
    import cPickle as pickle    
except ImportError:
    import pickle as pickle
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO
    
import cloudpickle
import pickledebug

class Serializer(object):
    
    serializedObject = None    
    _exception = None
    
    def __init__(self, obj):
        """Serialize a python object"""
        self.obj = obj
        
    def set_os_env_vars(self, os_env_vars):
        raise ValueError('Cannot transport os_env_vars on default serializer')
    
    def getexception(self):
        return self._exception
    
    def run_serialization(self, min_size_to_save=0):
        #min_size_to_save handled by subclass
        try:
            self.serializedObject = pickle.dumps(self.obj, protocol = 2)
            return self.serializedObject
        except pickle.PickleError, e:
            self._exception = e
            raise
        
    def get_module_dependencies(self): #can't resolve here..
        return []


class CloudSerializer(Serializer):
    """Use clould pickler"""

    _pickler = None
    _pickler_class = cloudpickle.CloudPickler
    os_env_vars = []
    
    def set_os_env_vars(self, os_env_vars):
        self.os_env_vars = os_env_vars
    
    def run_serialization(self, min_size_to_save=0):
    
        f = StringIO()
        
        self._pickler = self._pickler_class(f, protocol =2)
        self._pickler.os_env_vars = self.os_env_vars
        self.set_logged_object_minsize(min_size_to_save)
        
        try:
            self._pickler.dump(self.obj)            
            self.serializedObject = f.getvalue()
            return self.serializedObject
        except pickle.PickleError, e:
            self._exception = e
            raise
        
    def set_logged_object_minsize(self, minsize):
        #implemented by subclass
        pass
    
    def get_module_dependencies(self):
        return self._pickler.modules
    
class DebugSerializer(CloudSerializer):
    _pickler_class = pickledebug.CloudDebugPickler
            
    def write_debug_report(self, outfile,hideHeader=False):
        self._pickler.write_report(self.obj, outfile,hideHeader=hideHeader)
            
    def str_debug_report(self,hideHeader=False):
        """Get debug report as string"""    
        strfile = StringIO()
        self._pickler.write_report(self.obj, strfile,hideHeader=hideHeader)
        return strfile.getvalue()
    
    def set_report_minsize(self, minsize):
        self._pickler.printingMinSize = minsize
        
    def set_logged_object_minsize(self, minsize):
        self._pickler.min_size_to_save = minsize
            
                
class Deserializer(object):        
    deserializedObj = None
    
    def __init__(self, str):
        """Expects a python string as a pickled object which will be deserialized"""
        self.deserializedObj = pickle.loads(str)

    
        



import sys
import logging
import textwrap
from itertools import izip

try:
    import json
except ImportError:
    import simplejson as json
    
def safe_str(x):
    if isinstance(x, basestring):
        return x
    else:
        return str(x)    
    
def truncate(x):
    x = safe_str(x)
    x.strip()
    if '\n' in x:
        x, _ = x.split('\n', 2)
        x += ' ...'
    return x    

def safe_print(s):
    try:
        print s
    except UnicodeEncodeError:
        print s.encode(sys.stdout.encoding if sys.stdout.encoding else 'utf-8', errors='replace')

def safe_print_noline(s):
    try:
        print s,
    except UnicodeEncodeError:
        print s.encode(sys.stdout.encoding if sys.stdout.encoding else 'utf-8', errors='replace'),


class NullHandler(logging.Handler):
    """Does nothing with log messages. Included by default in the Python
    library for version 2.7+, but necessary for 2.6 compatibility"""
    
    def emit(self, record):
        pass

def list_of_dicts_printer(headers):
    """Prints a list of dictionaries as a table with an optional header"""
    
    def list_of_dicts_printer_helper(l, print_headers=True, kwargs={}):
        
        l = list(l)  # need to iterate twice through it
        spacing = {}
        for header in headers:
            max_column_width = len(header) if print_headers else 0
            for item in l:
                max_column_width = max(max_column_width, len(truncate(item.get(header, ''))))
            spacing[header] = max_column_width
        
        if print_headers:
            for header in headers:
                safe_print_noline( '%s%s'  % (header, ' ' * (spacing[header] - len(header) + 4)) )
            safe_print( '' )
        
        for item in l:
            for header in headers:
                safe_print_noline( '%s%s'  % (truncate(item.get(header, '')), ' ' * (spacing[header] - len(truncate(item.get(header, ''))) + 4)))
            print ''
    
    return list_of_dicts_printer_helper

def dict_printer(headers):
    """Prints a dictionary as a table with one row with an optional header"""
    
    def dict_printer_helper(d, print_headers=True, kwargs={}):
        list_of_dicts_printer(headers)([d], print_headers)
    
    return dict_printer_helper

def key_val_printer(key_header, value_headers):
    """Print a dictionary with keys mapping to value (or list of values).
    or dictionary of values. (if dictionary, value_headers should be superset of possible keys in value dictionary)    
    """
    
    headers = [key_header]
    
    if hasattr(value_headers, '__iter__'):
        headers.extend(value_headers)
        val_is_list = True
    else:
        headers.append(value_headers) 
        val_is_list = False
        
    def list_of_dicts_gen_singleval(d):
        """Where values of d are items"""
        return [{key_header : k, value_headers: v} for k, v in d.items()]
        
    def list_of_dicts_gen_list(d):
        """Where values of d are lists"""
        out_dcts = []
        for k, v in d.items():
            dct = dict(((vh, val) for vh, val in izip(value_headers, v)))
            dct[key_header] = k
            out_dcts.append(dct)
        return out_dcts              
    
    def list_of_dicts_gen_dict(d):
        """Where values of d are dictionaries"""
        out_dcts = []
        for k, v in d.items():
            dct = {key_header : k}
            dct.update(v)
            out_dcts.append(dct)
        return out_dcts              
        
    def helper(d, print_headers=True, kwargs={}):
        if val_is_list and isinstance(d.values()[0], dict):
            found_headers = set()
            for x in d.values():
                found_headers.update(x)
            return list_of_dicts_printer(found_headers)(list_of_dicts_gen_dict(d), print_headers)
                            
        lst_dct_gen = list_of_dicts_gen_list if val_is_list else list_of_dicts_gen_singleval
        return list_of_dicts_printer(headers)(lst_dct_gen(d), print_headers)
    
    return helper    
     

def list_printer(header):
    """Prints a list with one element per line with an optional header"""
    
    def list_printer_helper(l, print_headers=True, kwargs={}):
        
        if print_headers:
            safe_print(header)
        
        for item in l:
            safe_print(item)
        
        if getattr(l,'truncated', False):
            print >> sys.stderr, '...Results are truncated...'        
        
    return list_printer_helper

def volume_ls_printer(listings, print_headers=False, kwargs={}):
    """Prints the result of volume ls."""
    entries_printer = list_of_dicts_printer(['name', 'size', 'modified'])
    num_paths = len(listings)
    for i in range(num_paths):
        path, listing = listings[i]
        print path
        print 'total %s' % len(listing)
        entries_printer(listing, print_headers=False)
        if i < (num_paths - 1):
            print

    
            
def cloud_info_printer(info_results, print_headers, kwargs):
    """Dump contents"""
    
    base_ordering = ['status', 'exception', 'stdout', 'stderr', 'runtime', 
                     'created', 'finished', 'code_version', 'env', 'vol', 'exception', 'profile',
                
                ]
    info_ordering = kwargs.get('info_requested')
    if not info_ordering:
        info_ordering = base_ordering
    else:
        info_ordering = info_ordering.split(',')
        
    nl_info = ['stdout', 'stderr', 'exception', 'profile']
    
    started = False
    for jid, info_result in info_results.items():
        if started:
            print
        started = True
        
        # dyanamically update ordering as needed:
        for key in info_result:
            if key not in info_ordering:
                info_ordering.append(key)                
        
        print 'Info for jid %s' % jid                
        for info_key in info_ordering:
            result = info_result.get(info_key)
            if result:
                if info_key in nl_info:                
                    print '%s:' % info_key
                    safe_print(result)
                else:
                    safe_print('%s: %s' % (info_key, result))            

def bucket_info_printer(bucket_info_result, print_headers, kwargs):
    ordering = ['size', 'created', 'last-modified', 'md5sum', 'public', 'url']
    
    for key in bucket_info_result:
        if key not in ordering:
            ordering.append(key)                
                    
    for info_key in ordering:
        try:
            result = bucket_info_result[info_key]
        except KeyError:
            continue
        else:
            safe_print('%s: %s' % (info_key, result))            

def cloud_result_printer(results, print_headers, kwargs):
    started = False
    if len(results) == 1: # if just one result, print it directly (allows binary data to be written)
        val = results.values()[0]
        if isinstance(val, basestring):
            sys.stdout.write(val)
        else: # add newline if non-string
            print val 
    else:
        for jid, result in results.items():
            if started:
                print
            started = True
            print 'Result for jid %s:' % jid
            if isinstance(result, basestring):
                safe_print_noline(result)
            else: # add newline if non-string
                safe_print(result) 

    
def cloud_result_json_printer(results, print_headers, kwargs):
    """Values of results may already json encoded, hence the need for a special printer"""
    print '{',
    started = False 
    new_results = {}
    
    """First pass: Verify that all values are representable in json"""
    for jid, result in results.items():
        if not isinstance(result, basestring):
            result = json.dumps(result) # must be a primitive type at this stage
        elif not getattr(result, 'json_encoded', None): # functions.result marks results it encoded
            try:
                result = json.dumps(result)
            except (TypeError, UnicodeDecodeError):
                raise ValueError('result of jid %s cannot be represented in json. Please use default output format')
        new_results[jid] = result        
       
    for jid, result in new_results.items():
        if started:
            print ',',
        started = True
        safe_print_noline('"%s":' % jid)
        safe_print_noline('%s' % result)
            
    print '}'

def no_newline_printer(results, print_headers, kwargs):
    if results is None:
        pass
    else:
        print results,

from _abcoll import KeysView, ItemsView, ValuesView, MutableMapping

try:
    from thread import get_ident as _get_ident
except ImportError:
    from dummy_thread import get_ident as _get_ident



'''
For running commands in a shell environment on PiCloud.
Shell programs are executed through template substitution
'''
from __future__ import with_statement
from __future__ import absolute_import

"""
Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2011 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html    
"""

import os
import random
import re
import sys
import tempfile

from subprocess import Popen, PIPE

try:
    from json import dumps as json_serialize
except ImportError: #If python version < 2.6, we need to use simplejson
    from simplejson import dumps as json_serialize

from .util import template
from .rest import _low_level_publish
from .cron import _low_level_register
from .cloud import CloudException
from . import _getcloud

def _get_cloud_and_params(command, kwargs, ignore = []):
    for kwd in kwargs: 
        if not kwd.startswith('_'):
            raise ValueError('wildcard kwargs must be cloud kwd')        
    
    cloud = _getcloud()
    cloud._checkOpen()
    
    params = cloud._getJobParameters(None, kwargs, ignore)
    params['func_name'] = command
    params['fast_serialization'] = 2 # guarenteed to pass
    params['language'] = 'shell'
    
    return cloud, params    

def execute(command, argdict, return_file=None, ignore_exit_status=False, cwd=None, **kwargs):
    """Execute (possibly) templated *command*. Returns Job IDentifier (jid)
    
    * argdict - Dictionary mapping template parameters to values
    * return_file: Contents of this file will be result of job.  result is stdout if not provided
    * ignore_exit_status: if true, a non-zero exit code will not result in job erroring     
    * cwd: Current working directory to execute command within
    * kwargs: See cloud.call underscored keyword arguments
    
    """
    
    template.validate_command_args(command, argdict)
    
    _handle_args_upload(argdict)
    
    cloud, params = _get_cloud_and_params(command, kwargs)
    
    jid = cloud.adapter.job_call(params, _wrap_execute_program(command, return_file, ignore_exit_status, cwd = cwd), 
                                 (), argdict)
    return jid  

def execute_map(command, common_argdict, map_argdict, return_file=None, 
                ignore_exit_status=False, cwd=None, **kwargs):
    """Execute templated command in parallel. Return list of Job Identifiers (jids). See cloud.map
        for more information about mapping.  Arguments to this are:
        
    * common_argdict - Dictionary mapping template parameters to values for ALL map jobs
    * map_argdict - Dictionary mapping template parameters to a list of values
        The nth mapjob will have its template parameter substituted by the nth value in the list
        Note that all elements of map_argdict.values() must have the same length;
        The number of mapjobs produced will be equal to that length
    * return_file: Contents of this file will be result of job.  result is stdout if not provided
    * ignore_exit_status: if true, a non-zero exit code will not result in job erroring   
    * cwd: Current working directory to execute command within
    * kwargs: See cloud.map underscored keyword arguments     
    """
    #print 'c/m', common_argdict, map_argdict
    
    combined_dct = {}
    combined_dct.update(common_argdict)
    combined_dct.update(map_argdict)    
    template.validate_command_args(command, combined_dct)
    
    _handle_args_upload(common_argdict)
       
    # Convert map_argdict into a dist of dicts
    num_args = None
    
    map_dct_iters = {}
    # Error handling
    
    for key, val_list in map_argdict.items():
        if not num_args:
            num_args = len(val_list)
        if not val_list:
            raise ValueError('Key %s must map to a non-empty argument list' % key)
        elif num_args != len(val_list):
            raise ValueError('Key %s had %s args. Expected %s to conform to other keys' % (key, len(val_list), num_args))
        map_dct_iters[key] = iter(val_list)
        
    map_template_lists = [] # will be list of template dictionaries    
    
    if not num_args:
        raise ValueError('At least one element must be provided in map_argdict')
    
    for _ in xrange(num_args):
        map_template = {}
        for key, dct_iter in map_dct_iters.items():
            nxtval = next(dct_iter)
            map_template[key] = nxtval        
        
        _handle_args_upload(map_template)        
        map_template_lists.append(map_template)
    
    cloud, params = _get_cloud_and_params(command, kwargs)
        
    jids = cloud.adapter.jobs_map(params, 
                                 _wrap_execute_program(command, return_file, 
                                                       ignore_exit_status, common_argdict, cwd=cwd),                                 
                                None, map_template_lists)
    return jids


def rest_publish(command, label, return_file=None, 
                ignore_exit_status=False, **kwargs):
    """Publish shell *command* to PiCloud so it can be invoked through the PiCloud REST API
    The published function will be managed in the future by a unique (URL encoded) *label*.
    Returns url of published function. See cloud.rest.publish
    
    See cloud.shell.execute for description other arguments    
    See cloud.rest.publish for description of **kwargs
    """

    if not label:
        raise ValueError('label must be provided')
    m = re.match(r'^[A-Z0-9a-z_+-.]+$', label)
    if not m:
        raise TypeError('Label can only consist of valid URI characters (alphanumeric or from set(_+-.$)')
    try:
        label = label.decode('ascii').encode('ascii')
    except (UnicodeDecodeError, UnicodeEncodeError): #should not be possible
        raise TypeError('label must be an ASCII string')
    
    cloud, params = _get_cloud_and_params(command, kwargs,
                                          ignore=['_label', '_depends_on', '_depends_on_errors'] )
    
    # shell argspecs are dictionaries
    cmd_params = template.extract_vars(command)
    argspec = {'prms' : cmd_params,
               'cmd' : command}
    argspec_serialized = json_serialize(argspec)
    if len(argspec_serialized) >= 255: #won't fit in db - clear command
        del argspec['command']
        argspec_serialized = json_serialize(argspec)
        if len(argspec_serialized) >= 255: #commands too large; cannot type check
            argspec_serialized = json_serialize({})
    params['argspec'] = argspec_serialized
    
    return _low_level_publish(_wrap_execute_program(command, return_file, ignore_exit_status), 
                       label, 'raw', 'actiondct',
                       params, func_desc='command invoked in shell')['uri']

def cron_register(command, label, schedule, return_file = None,
                  ignore_exit_status=False, **kwargs):
    """Register shell *command* to be run periodically on PiCloud according to *schedule*
    The cron can be managed in the future by the specified *label*.
    
    Flags only relevant if you call cloud.result() on the cron job:
    return_file: Contents of this file will be result of job created by REST invoke.  
        result is stdout if not provided
    ignore_exit_status: if true, a non-zero exit code will not result in job erroring
    """

    cloud, params = _get_cloud_and_params(command, kwargs,
                                          ignore=['_label', '_depends_on', '_depends_on_errors'] )
    func = _wrap_execute_program(command, return_file, ignore_exit_status)
    
    return _low_level_register(func, label, schedule, params)
    
"""execution logic"""
def _execute_shell_program(command, return_file, ignore_exit_status, template_args, cwd = None):
    """Executes a shell program on the cloud"""
    
    _handle_args_download(template_args, cwd)    
    templated_cmd = template.generate_command(command, template_args)
    
    if not return_file: # must save commands stdout to a file                
        stdout_handle = PIPE
    else:
        stdout_handle = sys.stdout
        
    # ensure /home/scivm/ is present if any python interpreter is launched
    env = os.environ 
    cur_path = env.get('PYTHONPATH','')
    if cur_path:
        cur_path = ':%s'  % cur_path
    env['PYTHONPATH'] = '/home/scivm/' + cur_path
    
    #p = Popen(templated_cmd, shell=True, stdout=stdout_handle, stderr=PIPE, cwd=cwd, env=env)
    # execute in context of BASH for environment variables
    p = Popen(["/bin/bash", "-ic", templated_cmd], stdout=stdout_handle, stderr=PIPE, cwd=cwd, env=env)
    
    if stdout_handle == PIPE:
        # attach tee to direct stdout to file
        return_file = tempfile.mktemp('shellcmd_stdout')
        tee_cmd = 'tee %s' % return_file
        p_out = p.stdout
        tout = Popen(tee_cmd, shell=True, stdin=p_out, stdout=sys.stdout, stderr=sys.stderr, cwd=cwd)
    else:
        tout = None
    
    # capture stderr for exceptions
    stderr_file = tempfile.mktemp('shellcmd_stderr')
    tee_cmd = 'tee %s' % stderr_file
    p_err = p.stderr
    terr = Popen(tee_cmd, shell=True, stdin=p_err, stdout=sys.stderr, stderr=sys.stderr, cwd=cwd)
    
    retcode = p.wait()
    # give tee time to flush stdout
    terr.wait()
    if tout:  
        tout.wait()    
    
    if retcode:
        msg = 'command terminated with nonzero return code %s' % retcode
        if ignore_exit_status:
            print >> sys.stderr, msg
        else:
            msg += '\nstderr follows:\n'
            with open(stderr_file) as ferr:
                # ensure don't exceed storage limits
                ferr.seek(0,2)
                ferr_size = ferr.tell()                
                ferr.seek(max(0,ferr_size - 15000000), 0) 
                msg += ferr.read() 
            raise CloudException(msg)
    
    if cwd and not cwd.endswith('/'):
        cwd = cwd + '/'
    return_path = cwd + return_file if cwd and not return_file.startswith('/') else return_file
    
    try:
        with open(return_path,'rb') as f: # If this raises an exception, return file could not be read
            retval = f.read()
    except (IOError, OSError), e:
        if len(e.args) == 2:
            e.args = (e.args[0], e.args[1] + '\nCannot read return file!')
        raise    
    
    if stdout_handle == PIPE: 
        os.remove(return_file)
    os.remove(stderr_file)
        
    return retval

def _execute_program_unwrapper(command, return_file, ignore_exit_status, wrapped_args, template_args, cwd = None):
    """unwraps closure generated in _wrap_execute_program._execute_program_unwrapper_closure"""
    args = template_args
    if wrapped_args:
        args.update(wrapped_args)
    return _execute_shell_program(command, return_file, ignore_exit_status, args, cwd)

def _wrap_execute_program(command, return_file, ignore_exit_status, wrapped_args=None, cwd = None):
    """Used to put common arguments inside the stored function itself
    
    close over these arguments
    At execution, template_args are merged with wrapped_args
    """
    def _execute_program_unwrapper_closure(**template_args):
        """
        minimal function to avoid opcode differences between python2.6 and python2.7
        Code of this function is stored in pickle object; _execute_program_unwrapper is a global environment reference
        """
        return _execute_program_unwrapper(command, return_file, ignore_exit_status, 
                                          wrapped_args, template_args, cwd)
    
    return _execute_program_unwrapper_closure 

"""helper functions"""
"""File uploading logic

There are some inefficiencies here.
    By closing the file data in a function, we lose the ability to stream it from disk 
    In practical usage, this probably won't matter and can always be changed later
        by using a rest invoke interface

"""

action_default = 'action_default'
action_upload = 'action_upload'

def _encode_upload_action(file_name):
    # upload file by binding it to a closure    
    
    f = open(file_name,'rb')
    contents = f.read()
    f.close()
    
    base_name = os.path.basename(file_name)
    
    return {'action' : action_upload,
            'filename' : base_name,
            'contents' : contents}
    
def _encode_default_action(arg):
    return {'action' : action_default,
            'value' : arg}    

def _handle_args_upload(arg_dct):
    """"
    arg_dct is a dictionary describing a job
    Each key is parameter that maps to its argument value
    
    If an argument is a file, it is automatically replaced by a function 
    that handles file unpacking
    """
    
    for param, arg in arg_dct.items():
        if arg.startswith('@'): # defines a file
            arg_dct[param] = _encode_upload_action(arg[1:])
        else:
            arg_dct[param] = _encode_default_action(arg)
    
    
"""downloading"""

def _decode_upload_action(action_dct, cwd):
    """place data in the current directory
    file name is name 
    if name already exists, append random integers to name until it doesn't
    """
    
    name = action_dct['filename']
    contents = action_dct['contents']
    
    cloud = _getcloud()
    if not cloud.running_on_cloud(): # simulation
        name = tempfile.mktemp(suffix=name)        
    
    started = False
    while os.path.exists(name):
        if not started:
            name+='-'
        started = True
        name += str(random.randint(0,9))
    
    if cwd and not cwd.endswith('/'):
        cwd = cwd + '/'
        
    fullpath = cwd + name if cwd else name 
    
    # Write user-uploaded file to local storage. (Can fail due to permission issues)
    # Be sure it has executable permissions on incase it is a shell script
    f = os.fdopen(os.open(fullpath,os.O_CREAT|os.O_RDWR,0777),'wb')
    f.write(contents)
    f.close()
    
    return name # use local name to fill in template

def _decode_default_action(action_dct, cwd):
    return action_dct['value']

def _handle_args_download(arg_dct, cwd):
    decode_map = {
                  action_upload : _decode_upload_action,
                  action_default : _decode_default_action 
                  }
    
    for param, action_dct in arg_dct.items():
        arg_dct[param] = decode_map[action_dct['action']](action_dct, cwd)



"""
Classes are responsible for streaming compression of objects

Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2010 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

import struct
from cStringIO import StringIO

from .gzip_stream import GzipFile


class Packer(object):
    """Object packer"""
    
    def __init__(self, output = None, compresslevel=5):
        if not output:
            output = StringIO()
        self._output = output
        self._gz = GzipFile(compresslevel=compresslevel, fileobj = self._output, mode='wb')
        
    def add(self, str_obj):
        """Add a string object into the packer"""
        
        sz = struct.pack('!I',len(str_obj)) #write size
        self._gz.write(sz)
        self._gz.write(str_obj)

    def finish(self):
        """Close compressor and return output"""
        self._gz.close()
        return self._output
    
class UnPacker(object):
    """Unpacks the format that Packer generates
    Expects a file-like obj as its first argument
    Acts as an iterator"""
    
    def __init__(self, file_obj):
        self._gz = GzipFile(fileobj = file_obj, mode = 'rb')
        
    def __iter__(self):
        return self
    
    def next(self):
        if not self._gz:
            raise StopIteration
        size_str = self._gz.read(4)        
        if not size_str:
            self._gz.close()
            self._gz = None
            raise StopIteration
        sz = struct.unpack('!I', size_str)[0]        
        retval = self._gz.read(sz)
        if len(retval) < sz:
            self._gz.close()
            self._gz = None
            raise IOError('file obj seems truncated')
        return retval

def test():
    """unit test code"""
    
    v = '123'*200
    w = '456'
    x = '789'
    
    p = Packer()
    p.add(v)
    p.add(w)
    p.add(x)
    
    packed = p.finish()    
    
    packed.seek(0)
    print 'len %d' % len(packed.getvalue())
    print packed.getvalue()
    
    
    for val in UnPacker(packed):
        print 'val is %s' % val



"""
Science VM environment management.
This module allows the user to manage their environments on Science VM.
See documentation at http://docs.scivm.com
"""
from __future__ import absolute_import
"""
Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2013 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see
http://www.gnu.org/licenses/lgpl-2.1.html
"""
try:
    import json
except:
    # Python 2.5 compatibility
    import simplejson as json
import logging
import platform
import random
import re
import sys
import string
import time

import scicloud as cloud
from .cloudlog import stdout as print_stdout, stderr as print_stderr
from .util import credentials
from .util import common

cloudLog = logging.getLogger('Cloud.environment')

plat = platform.system()

_urls = {'list': 'environment/list/',
         'list_bases': 'environment/list_bases/',
         'create': 'environment/create/',
         'edit_info': 'environment/edit_info/',
         'modify': 'environment/modify/',
         'save': 'environment/save/',
         'save_shutdown': 'environment/save_shutdown/',
         'shutdown': 'environment/shutdown/',
         'clone': 'environment/clone/',
         'delete': 'environment/delete/',
         }

# environment status types
_STATUS_CREATING = 'new'
_STATUS_READY = 'ready'
_STATUS_ERROR = 'error'

# environment action types
_ACTION_IDLE = 'idle'
_ACTION_SETUP = 'setup'
_ACTION_EDIT = 'edit'
_ACTION_SAVE = 'save'
_ACTION_SETUP_ERROR = 'setup_error'
_ACTION_SAVE_ERROR = 'save_error'


def _send_env_request(request_type, data, jsonize_values=True):
    type_url = _urls.get(request_type)
    if type_url is None:
        raise LookupError('Invalid env request type %s' % request_type)
    return common._send_request(type_url, data, jsonize_values)

"""
environment management
"""
def list_envs(name=None):
    """Returns a list of dictionaries describing user's environments.
    If *name* is given, only shows info for the environment with that name.

    Environment information is returned as list of dictionaries.  The keys
    within each returned dictionary are:

    * name: name of the environment
    * status: status of the environment
    * action: the action state of the environment (e.g. under edit)
    * created: time when the environment was created
    * last_modifed: last time a modification was saved
    * hostname: hostname of setup server if being modified
    * setup_time: time setup server has been up if being modified
    """
    resp = _send_env_request('list', {'env_name': name})
    return [common._fix_time_element(env, ['created', 'last_modified'])
            for env in resp['envs_list']]

def list_bases():
    """Returns a list of dictionaries describing available bases. The keys
    within each returned dictionary are:

    * id: id of the base (to be used when referencing bases in other functions)
    * name: brief descriptive name of the base
    """
    resp = _send_env_request('list_bases', {})
    return resp['bases_list']

def create(name, base, desc=None):
    """Creates a new cloud environment.

    * name: name of the new environment (max 30 chars)
    * base: name of the base OS to use for the environment (use list_bases to
        see list of bases and their names)
    * desc: Optional description of the environment (max 2000 chars)

    Returns the hostname of the setup server where the newly created
    environment can be modified.
    """
    pattern = '^[a-zA-Z0-9_-]*$'
    if not name:
        raise cloud.CloudException('No environment name given')
    elif len(name) > 30:
        raise cloud.CloudException('Environment name cannot be more than 30'
                                   ' characters')
    elif not re.match(pattern, name):
        raise cloud.CloudException('Environment name must consist of letters,'
                                   ' numbers, underscores, or hyphens')
    if desc and len(desc) > 2000:
        raise cloud.CloudException('Environment description cannot be more'
                                   ' than 2000 characters')

    resp = _send_env_request('create',
                             {'env_name': name, 'base_name': base,
                              'env_desc': desc or ''})
    cloudLog.debug('created environment %s', resp['env_name'])
    return get_setup_hostname(name)

def edit_info(name, new_name=None, new_desc=None):
    """Edits name and description of an existing environment.

    * name: current name of the environment
    * new_name: Optional new name of the environment (max 30 chars)
    * new_desc: Optional new description of the environment (max 2000 chars)
    """
    if new_name is None and new_desc is None:
        return

    pattern = '^[a-zA-Z0-9_-]*$'
    if not name:
        raise cloud.CloudException('No environment name given')

    if new_name is not None:
        if len(new_name) > 30:
            raise cloud.CloudException('Environment name cannot be more than 30'
                                       ' characters')
        elif not re.match(pattern, name):
            raise cloud.CloudException('Environment name must consist of letters,'
                                       ' numbers, underscores, or hyphens')
    if new_desc is not None and len(new_desc) > 2000:
        raise cloud.CloudException('Environment description cannot be more'
                                   ' than 2000 characters')

    resp = _send_env_request('edit_info',
                             {'old_env_name': name, 'new_name': new_name,
                              'new_desc': new_desc})
    cloudLog.debug('edited info for environment %s', resp['env_name'])

def modify(name):
    """Modifies an existing environment.

    * name: name of environment to modify

    Returns the hostname of the setup server where environment can be modified.
    """
    resp = _send_env_request('modify', {'env_name': name})
    cloudLog.debug('modify requested for env %s', resp['env_name'])
    return get_setup_hostname(name)

def save(name):
    """Saves the current modified version of the environment, without tearing
    down the setup server.

    * name: name of the environment to save

    This is a blocking function.  When it returns without errors, the new
    version of the environment is available for use by all workers.
    """
    resp = _send_env_request('save', {'env_name': name})
    cloudLog.debug('save requested for env %s', resp['env_name'])
    wait_for_edit(name)

def save_shutdown(name):
    """Saves the current modified version of the environment, and tears down
    the setup server when saving is done.

    * name: name of the environment to save

    This is a blocking function.  When it returns without errors, the new
    version of the environment is available for use by all workers.
    """
    resp = _send_env_request('save_shutdown', {'env_name': name})
    cloudLog.debug('save_shutdown requested for env %s', resp['env_name'])
    wait_for_idle(name)

def shutdown(name):
    """Tears down the setup server without saving the environment modification.

    * name: name of the environment to save
    """
    resp = _send_env_request('shutdown', {'env_name': name})
    cloudLog.debug('shutdown requested for env %s', resp['env_name'])
    wait_for_idle(name)

def clone(parent_name, new_name=None, new_desc=None):
    """Creates a new cloud environment by cloning an existing one.

    * parent_name: name of the existing environment to clone
    * new_name: new name of the environment. default is
        parent_name + "_clone". (max 30 chars)
    * new_desc: Optional description of the environment if different from
        parent environment description. (max 2000 chars)
    """
    pattern = '^[a-zA-Z0-9_-]*$'
    new_name = new_name or (parent_name + '_clone')
    if len(new_name) > 30:
        raise cloud.CloudException('Environment name cannot be more than 30'
                                   ' characters')
    elif not re.match(pattern, new_name):
        raise cloud.CloudException('Environment name must consist of letters,'
                                   ' numbers, underscores, or hyphens')
    if new_desc and len(new_desc) > 2000:
        raise cloud.CloudException('Environment description cannot be more'
                                   ' than 2000 characters')

    resp = _send_env_request('create',
                             {'parent_env_name': parent_name,
                              'env_name': new_name,
                              'env_desc': new_desc})
    cloudLog.debug('created environment %s', resp['env_name'])
    wait_for_idle(new_name)

def delete(name):
    """Deletes and existing environment.

    * name: Name of the environment to save
    """
    resp = _send_env_request('delete', {'env_name': name})
    cloudLog.debug('delete requested for env %s', resp['env_name'])

def get_setup_hostname(name):
    """Returns the hostname of the setup server where environment can be
    modified.  raises exception if the environment does not have a setup server
    already launched.

    * name: name of the environment whose setup server hostname is desired
    """
    env_info = wait_for_edit(name, _ACTION_IDLE)
    if env_info is None:
        raise cloud.CloudException('Environment is not being modified')
    return env_info['hostname']

def get_key_path():
    """Return the key file path for sshing into setup server."""
    api_key = cloud.connection_info().get('api_key')
    return credentials.get_sshkey_path(api_key)

def ssh(name, cmd=None):
    """Creates an ssh session to the environment setup server.

    * name: Name of the environment to make an ssh connection
    * cmd: By default, this function creates an interactive ssh session.
        If cmd is given, however, it executes the cmd on the setup server
        and returns the output of the command execution.
    """
    hostname = get_setup_hostname(name)
    key_path = get_key_path()
    status, stdout, stderr = common.ssh_session('scivm', hostname, key_path,
                                                run_cmd=cmd)
    if status:
        if stdout:
            sys.stdout.write(stdout)
        if stderr:
            sys.stderr.write(stderr)
        sys.exit(status)

    if cmd and stdout:
        return stdout

def rsync(src_path, dest_path, delete=False, pipe_output=False):
    """Syncs data between a custom environment and the local filesystem. A
    setup server for the environment must already be launched. Also, keep in
    mind that the scivm user account (which is used for the rsync operation)
    has write permissions only to the home directory and /tmp on the setup
    server. If additional permissions are required, consider doing the rsync
    manually from the setup server using sudo, or rsync to the home directory
    then do a subsequent move using sudo.

    Either *src_path* or *dest_path* should specify an environment path, but
    not both. An environment path is of the format:

        env_name:[path-within-environment]

    Note that the colon is what indicates this is an environment path
    specification.

    *src_path* can be a list of paths, all of which should either be local
    paths, or environment paths. If *src_path* is a directory, a trailing slash
    indicates that its contents should be rsynced, while ommission of slash
    would lead to the directory itself being rsynced to the environment.

    Example::

        rsync('~/dataset1', 'my_env:')

    will ensure that a directory named 'dataset1' will exist in the user
    scivm's home directory of environment 'my_env'. On the other hand,

        rsync(~/dataset1/', 'my_env:')

    will copy all the contents of 'dataset1' to the home directory of user
    scivm. See rsync manual for more information.

    If *delete* is True, files that exist in *dest_path* but not in *src_path*
    will be deleted.  By default, such files will not be removed.
    """
    conn = cloud._getcloudnetconnection()
    retry_attempts = conn.retry_attempts
    dest_is_local = common.is_local_path(dest_path)
    l_paths, r_paths = ((dest_path, src_path) if dest_is_local else
                        (src_path, dest_path))
    local_paths = common.parse_local_paths(l_paths)
    env_name, env_paths = common.parse_remote_paths(r_paths)

    hostname = get_setup_hostname(env_name)
    try:
        r_base = 'scivm@%s:' % hostname
        r_paths = ' '.join(['%s%s' % (r_base, path) for path in env_paths])
        l_paths = ' '.join(local_paths)
        sync_args = (r_paths, l_paths) if dest_is_local else (l_paths, r_paths)

        for attempt in xrange(retry_attempts):
            exit_code, _, _ = common.rsync_session(*sync_args, delete=delete,
                                                   pipe_output=pipe_output)
            if not exit_code:
                break
            print_stderr('Retrying environment rsync...')
        else:
            raise Exception('rsync failed multiple attempts... '
                            'Please contact PiCloud support')
    except Exception as e:
        cloudLog.error('Environment rsync errored with:\n%s', e)
        print e

def run_script(name, filename):
    """Runs a script on the environment setup server, and returns the output.

    * name: Environment whose setup server should run the script
    filename: local path where the script to be run can be found
    """
    POPU = string.ascii_letters + string.digits
    dest_file = ''.join(random.sample(POPU, 16))
    try:
        rsync(filename, '%s:%s' % (name, dest_file), pipe_output=True)
        run = "chmod 700 {0}; ./{0} &> {0}.out; cat {0}.out".format(dest_file)
        output = ssh(name, run)
    except Exception as e:
        cloudLog.error('script could not be run: %s', str(e))
        print 'Script could not be run on the setup server.'
        print e
    else:
        return output
    finally:
        ssh(name, "rm -rf %s*" % dest_file)

def wait_for_idle(name, invalid_actions=None):
    """Waits for environment to be in idle action state."""
    return _wait_for(name=name, action=_ACTION_IDLE,
                     invalid_actions=invalid_actions)

def wait_for_edit(name, invalid_actions=None):
    """Waits for environment to be in edit action state."""
    return _wait_for(name=name, action=_ACTION_EDIT,
                     invalid_actions=invalid_actions)

def _wait_for(name, action, invalid_actions=None, poll_frequency=2,
              max_poll_duration=1800):
    """Generic wait function for polling until environment reaches the
    specified action state. Raises exception if the environment ever falls into
    an error status or action state.
    """
    invalid_actions = invalid_actions or []
    if not hasattr(invalid_actions, '__iter__'):
        invalid_actions = [invalid_actions]

    for _ in xrange(max_poll_duration / poll_frequency):
        resp = list_envs(name)

        if len(resp) == 0:
            raise cloud.CloudException('No matching environment found.')
        elif len(resp) != 1:
            cloudLog.error('single env query returned %s results', len(resp))
            raise cloud.CloudException('Unexpected result from PiCloud. '
                                       'Please contact PiCloud support.')
        env_info = resp.pop()
        resp_status = env_info['status']
        resp_action = env_info['action']
        if resp_status == _STATUS_ERROR:
            raise cloud.CloudException('Environment creation failed. '
                                       'Please contact PiCloud support.')
        elif resp_status == _STATUS_READY:
            if resp_action == _ACTION_SETUP_ERROR:
                raise cloud.CloudException('Setup server launch failed. '
                                           'Please contact PiCloud support.')
            elif resp_action == _ACTION_SAVE_ERROR:
                raise cloud.CloudException('Environment save failed. '
                                           'Please contact PiCloud support.')
            elif resp_action in invalid_actions:
                return None
            elif resp_status == _STATUS_READY and action == resp_action:
                return env_info
        elif resp_status == _STATUS_CREATING:
            pass

        time.sleep(poll_frequency)

    raise cloud.CloudException('Environment operation timed out. '
                               'Please contact PiCloud support.')



"""
This implements the multiprocessing adapter.  Much is added to the baseline Multiprocessing.Pool

Note: This code is unfortunately a bit of a mess due to the sheer number of extensions
Ideally, it would be rewritten one day

Clean-up notes:
    -Seperate classes into various modules
    -Clear-up some duplicate code (killing is very messy esp w/ process respawn 
        - borrow code from pool - 2.7)
        - clear_results is rediculous -- should interface w/ handle_results (in pool)
    -Consider no longer subclassing from pool -- only leads to more confusion w/ logging, etc
    -Seperate open into variety of functions
    

Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2011 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

from __future__ import with_statement
import sys
import os
import inspect
import time
import types
import socket
import traceback
import threading
import itertools
import Queue
import traceback
import cProfile
import signal
from multiprocessing import cpu_count, current_process
from multiprocessing.managers import BaseManager, State, Server, Token, AutoProxy, BaseProxy, dispatch
from multiprocessing.pool import Pool
import multiprocessing.pool as pool
from multiprocessing.util import Finalize, debug, _run_finalizers
#todo: Set debug to use cloudLog?

from ..cloud import CloudException, CloudTimeoutError
from ..serialization import deserialize
from ..util.xrange_helper import maybe_xrange_iter
from .connection import CloudConnection
from .adapter import SerializingAdapter
from .. import cloudconfig as cc
import logging
cloudLog = logging.getLogger('Cloud.mp') 

use_profiler = False   #For internal testing use

#status:
#Note: No assigned as it makes no sense in this context
status_waiting = 'waiting'  #waiting for dep to be satisfied
status_new = 'queued'  #new job in queue  (should this be queued??)
status_processing = 'processing' #processing
status_done = 'done' #job done
status_error = 'error' #job messed up
status_stalled = 'stalled' #job fails due to dependency failing
status_killed = 'killed' #job explicitly killed

ran_statuses = [status_done, status_error, status_stalled, status_killed]
error_statuses = [status_error, status_stalled, status_killed]

#constants
start_processing_flag = True
sleep_task_flag = True

#timer taken from timeit
if sys.platform == "win32":
    # On Windows, the best timer is time.clock()
    default_timer = time.clock
else:
    # On most other platforms the best timer is time.time()
    default_timer = time.time


class KilledException(Exception):
    """Exception used when job killed"""
    pass

#
# Code run by worker processes
# Modified from base mp to notify when starting to process and receive backoff messages
#
def cloud_worker(inqueue, outqueue, cloudHooks, mypid = 0, redirect_job_output=True, 
                 serialize_debugging = False, serialize_logging = False,
                 api_key=None, api_secretkey=None):
    put = outqueue.put
    get = inqueue.get
    if hasattr(inqueue, '_writer'):
        inqueue._writer.close()
        outqueue._reader.close()

    #bind last keys:
    from .. import cloudinterface
    cloudinterface.last_api_key = api_key
    cloudinterface.last_api_secretkey = api_secretkey
    
    new_cloud = patch_cloud(cloudHooks, mypid)
    adapter = new_cloud.adapter
    report = adapter.report   
    
    if not report and redirect_job_output:  #disable redirection!
        cloudLog.warn('Cannot redirect Job Output due to report construction failing. Falling back to console output')
        redirect_job_output = False 
    
    #options:
    adapter.serializeDebugging = serialize_debugging
    adapter.serializeLogging = serialize_logging

    base_stdout = sys.stdout
    base_stderr = sys.stderr
        
    #print 'worker online'
    
    while 1:
        try:
            task = get()
            #print 'worker got', task
        except (EOFError, IOError):
            debug('worker got EOFError or IOError -- exiting')
            break

        if task is None:
            debug('worker got sentinel -- exiting')
            break
        
        if task is sleep_task_flag: #help killing
            time.sleep(2.0)
            continue

        job, func, args, kwds, result_serialization_level, logname, logpid, logcnt = task
        
        if func:
            func = deserialize(func)
        if args:
            args = deserialize(args)
        if kwds:
            kwds = deserialize(kwds)
        
        put ((job, mypid, start_processing_flag)) #None as result indicates this is started
                        
        startTime = default_timer() #in case exception triggers early
        try:                                                                        
            if redirect_job_output:  #need a proxy file writer to redirect!
                stdout = report.open_report_file(logname, '.stdout', logcnt, logpid)
                stderr = report.open_report_file(logname, '.stderr', logcnt, logpid) 
                
                sys.stdout = stdout 
                sys.stderr = stderr
            
            funcres = None
            
            # catch SIGINT (kill commands) and simply raise an exception
            # so that we get access to the traceback            
            # Note: os.kill() doesn't work on win32 for python < 2.7, preventing signaling
            if sys.platform != 'win32' or sys.version_info >= (2.7):
                signal.signal(signal.SIGINT, sigterm_handler) 
            
            if use_profiler and report:
                pfname = report.get_report_file(logname, '.profile', logcnt, logpid)
                #print pfname
                
                prof = cProfile.Profile()
                startTime = default_timer()
                funcres = prof.runcall(func, *args, **kwds)
                prof.dump_stats(pfname)
            else:
                startTime = default_timer()
                funcres = func(*args, **kwds)  
            endTime = default_timer()     
            
            sfuncres = adapter.getserializer(result_serialization_level)(funcres)  
            try:
                sfuncres.run_serialization()          
            finally:
                if adapter.serializeLogging and report:                    
                    acname = report.save_report(sfuncres, logname, logcnt, logpid)
                else:
                    acname = ''
            #log here + checksize
            adapter.check_size(sfuncres, acname)
                        
            result = (True, sfuncres.serializedObject)
        except (SystemExit, Exception), e: 
            endTime = default_timer()
            tb = traceback.format_exc()       
            excp_type = type(e)
            # hack: Sometimes custom tyoes aren't serializable. So just fallback to Exception for safety 
            if excp_type.__module__ not in ['exceptions', 'cloud.transport.local']:
                excp_type = Exception                
            result = (excp_type, tb)
        finally:
            # set signal handler back to default
            if sys.platform != 'win32' or sys.version_info >= (2.7):
                signal.signal(signal.SIGINT, signal.SIG_DFL)                      
            
            if redirect_job_output:                
                if sys.stdout != base_stdout:
                    sys.stdout.close()
                    sys.stdout = base_stdout    
                if sys.stdout != base_stderr:
                    sys.stderr.close()
                    sys.stderr = base_stderr
            
        put((job, endTime - startTime, result))

def sigterm_handler(signum, frame):
    """Handler for preserving tracebacks on cloud.kill"""
    cloudLog.info('Signal handler called with signal %s', signum)    
    signal.signal(signal.SIGINT, signal.SIG_DFL)  #back to default                       
    raise KilledException('Job terminated by user')


handlerInit = False
def setupLogHandlers():
    """Debug statements to bind multiprocessing to cloudlog"""        
    
    global handlerInit
    if handlerInit:
        return
    handlers = cloudLog.parent.handlers #found in main cloudlog
    
    from multiprocessing import get_logger
    
    mplogger = get_logger()
   
    mplogger.setLevel(3) #bind to cloudlog
       
    for handler in handlers:
        mplogger.addHandler(handler)
    
    handlerInit = True
        
class CloudJob(object):
    """CloudJob: 
        Modified version of ApplyResult.
    """
    killMe = False #set to true to self-destruct once processing notification comes in
    deleted = False #used only for killing deleted jobs - if set, do not flush result back 
    _runningpid = None
    _runtime = 0.0 #default runtime
    
    job_counter = itertools.count() #class counter
    job_counter.next() #coerce to start at 2

    
    def __init__(self, pool, func, args, kwds, func_name, label, priority,
                 result_serialization_level, logname="", logpid = 1, logcnt=0):
        self._jid = self.job_counter.next()
        self._pool = pool        
        pool.cache[self._jid] = self
        self._status = status_new
        self._value = None
        self.func = func
        self.args = args
        self.kwds = kwds
        self.func_name = func_name
        self.label = label
        self.priority = priority
        self.result_serialization_level = result_serialization_level
        self.logname = logname
        self.logpid = logpid
        self.logcnt = logcnt        

    @property
    def ready(self):        
        return self._status in ran_statuses 
    
    @property
    def jid(self):
        return self._jid
    
    @property
    def status(self):
        return self._status
    
    @property
    def pid(self):
        return self._pid
    
    @property
    def result(self):
        return self._value
    
    @property
    def runtime(self):
        st = getattr(self, '_start_time', None)
        if st:
            return default_timer() - st
        else:
            return self._runtime
    
    def _aborted(self, abort_status, result = None):
        if hasattr(self, '_start_time'):
            endTime = time.time()
            self._runtime = endTime - self._start_time
            del self._start_time        
        self._status = abort_status
        self._value = result        
        self._pool.notify_job_done(self._jid)
        try:
            self._pool._task_sem.release()
        except ValueError, e: 
            cloudLog.debug('Semaphore release triggered %s', str(e))
        del self._pool.cache[self._jid]
    
    def _set(self, extra_data, sobj):
        """Update with some arbitrary extra_data and some representation of a returned object"""
        if sobj is start_processing_flag: #flag for processing
            #print 'set to processing %s k=? %s' % (self.status, self.killMe)
            if self.status == status_killed: #killed?
                return             
            self._start_time = default_timer()  
            self._pid = extra_data             
            
            self._status = status_processing
            
            
            try:
                self._pool._task_sem.release()
            except ValueError, e: 
                cloudLog.debug('Semaphore release triggered %s', str(e))

            
            if self.killMe: #important: this check must come AFTER processing set
                cloudLog.debug("Auto-Killing job after started processing. jid=%d, label=%s, func=%s, " % (self.jid, self.label, self.func_name))
                self._pool._do_job_kill(self)
            else:
                cloudLog.debug("Processing job. jid=%d, label=%s, func=%s, " % (self.jid, self.label, self.func_name))            
            return                               
        
        if self.status in ran_statuses: #killed?
            return
        
        self._runtime = extra_data
        del self._start_time
        success, self._value = sobj        
        if success == True:
            self._status = status_done
            self._value = self._value
        elif self._status != status_killed:
            if success == KilledException:
                self._status = status_killed
            else:
                self._status = status_error
            

        cloudLog.debug("Finished job. jid=%d, label=%s, func=%s, status=%s, runtime=%g" % (self.jid, self.label, self.func_name, self.status,  self.runtime))            
        
        self._pool.notify_job_done(self._jid)
        
        try: #might not exist if killing
            del self._pool.cache[self._jid]
        except:
            pass
        
    def __str__(self):
        return 'CloudJob <jid: '+str(self.jid)+', func_name: '+self.func_name + \
                         ', label: '+str(self.label)+', status:'+self.status + '>'

class ThreadedServer(Server):
    """With self descruct ability"""
    
    destroy_self = False
    close_at = None #if set to a time, break at this time
    close_wait_time = 0.1 #time to wait for closing
     
    def serve_forever(self):
        '''
        Run the server forever
        '''
        current_process()._manager_server = self
        try:
            try:
                while 1:
                    try:
                        c = self.listener.accept()
                    except socket.timeout:
                        cloudLog.debug('Timeout triggered')
                    except (OSError, IOError):
                        continue
                    if self.destroy_self:
                        if self.close_at:
                            if time.time() > self.close_at: 
                                break
                        else:                                                        
                            self.close_at = time.time() + self.close_wait_time
                            #massive hack: allow accept to abort with a timeout
                            self.listener._listener._socket.settimeout(self.close_wait_time)                            
                            cloudLog.debug('Threaded server will shut down at %s', self.close_at)
                    t = threading.Thread(target=self.handle_request, args=(c,), name='handle_request')
                    t.daemon = True
                    t.start()
            except (KeyboardInterrupt, SystemExit):
                pass
        finally:
            self.stop = 999
            self.listener.close()
            cloudLog.info('ThreadedServer shut down. destroy_self was %s' % self.destroy_self)
           
class MPManager(BaseManager):
    """Modified to run in same thread as initializer
    IMPORTANT: Because this binds the instance to the class, only one of these objects can ever exist"""
    server = None
    
    waitCV = None
    
    myInstance = None #only once instance allowed
    
    _Server = ThreadedServer

    @classmethod
    def __new__(cls, *args, **kwds):
        """Singleton -- Do not allow multiple initializations"""
        if cls.myInstance:
            return None        
        else: 
            instance =  BaseManager.__new__(*args, **kwds)
            cls.myInstance = instance
            return instance

    @classmethod
    def _run_server_nosend(cls, registry, address, authkey, serializer):
        '''
        Create a server, report its address and run it
        '''
        
        # create server
        cls.server = cls._Server(registry, address, authkey, serializer)

        # run the manager
        cloudLog.debug('Mpmanager serving at %r', cls.server.address)
               
        with cls.waitCV:
            cls.waitCV.notify()
        #print 'try start server..'
        cls.server.serve_forever() 
           
    
    def start_threaded(self):
        """Launch this as a threaded application"""
        assert self._state.value == State.INITIAL

        self.__class__.waitCV = threading.Condition()
        
        # spawn thread which runs the server
        self._thread = threading.Thread(
            target=type(self)._run_server_nosend,
            name = 'run_server_nosend',
            args=(self._registry, self._address, self._authkey,
                  self._serializer)
            )
        self._thread.daemon = True

        
        with self.__class__.waitCV:
            self._thread.start()
            self.__class__.waitCV.wait()

        ident = self._thread.ident if hasattr(self._thread,'ident') else '' #python 2.5 lacks ident
        cloudLog.debug('Manager started with thread ident %s' % ident)
        self._thread.name = type(self).__name__  + '-' + str(ident)
        self._address = self.__class__.server.address  
        
        # register a finalizer (TODO):
        self._state.value = State.STARTED
        
    def close(self):   
        """
        Closing is tricky, due to risk of receiving on a dead connection
        Strategy we use is to force mpmanager to wait a little before fully closing"""
        
        cloudLog.debug('Starting pool close')
        self.__class__.server.destroy_self = True            
        
        try:
            conn = self._Client(self._address) #do not attempt to receive data as connection may be dead
        except (IOError, EOFError): #all good
            if not self._thread.is_alive():                
                cloudLog.debug('MpManager successfully shut down early')
            else:    
                cloudLog.warning('MPManager Thread alive but cannot bind to %s', self._address, exc_info=True)                    
        except Exception:
            cloudLog.debug('Cloud not bind to %s.' % self._address, exc_info=True)
        else: 
            try:                
                cloudLog.debug('Client to server shutdown %s' % self._address)
                conn.send(('#RETURN', None)) #garbage to force server to self-descruct
                cloudLog.debug('Sent #RETURN')
            finally:                
                conn.close()            
        
        cloudLog.debug('Finalizers start')
        _run_finalizers()
        cloudLog.debug('Finalizers end -- killing manager')
        
            
        #block until mpmanager down
        self._thread.join(self.__class__.server.close_wait_time + 0.07)
        if self._thread.is_alive():
            cloudLog.error('Failed to shut down mp manager!')
        else:
            cloudLog.debug('mp manager shut down successfully')
            
        #reset!
        self.__class__.myInstance = None
        try:
            del BaseProxy._address_to_local[self._address]
        except KeyError: #may already be removed by finalizer
            pass
        current_process()._tempdir = None 
            
    
    @classmethod
    def register(cls, typeid, callable=None, proxytype=None, exposed=None,
                 method_to_typeid=None, create_method=True):
        '''
        Register a typeid with the manager type
        Slight modification to in the method created
        '''
        if '_registry' not in cls.__dict__:
            cls._registry = cls._registry.copy()

        if proxytype is None:
            proxytype = AutoProxy

        exposed = exposed or getattr(proxytype, '_exposed_', None)

        method_to_typeid = method_to_typeid or \
                           getattr(proxytype, '_method_to_typeid_', None)

        if method_to_typeid:
            for key, value in method_to_typeid.items():
                assert type(key) is str, '%r is not a string' % key
                assert type(value) is str, '%r is not a string' % value

        cls._registry[typeid] = (
            callable, exposed, method_to_typeid, proxytype
            )

        if create_method:
            def temp(self, *args, **kwds):
                debug('requesting creation of a shared %r object', typeid)

                oid, exp = self.__class__.server.create(None,typeid, *args, **kwds)
                token = Token(typeid, self._address, oid)
                proxy = proxytype(
                    token, self._serializer, manager=self,
                    authkey=self._authkey, exposed=exp
                    )
                conn = self._Client(token.address, authkey=self._authkey)
                dispatch(conn, None, 'decref', (token.id,))
                return proxy
            temp.__name__ = typeid
            setattr(cls, typeid, temp)        

class MPConnectionHook(object):
    """Wrapper object to make calls to MPConnection from subprocesses
    """
    
    def __init__(self, cloudcon, cloudmodule, reportDir):       
        self.cloudcon = cloudcon
        self.cloudmodule = cloudmodule
        self.reportDir = reportDir
    
    def open(self):
        """Open master if it isn't open"""
        if not self.cloudcon.opened:
            self.cloudcon.open()
        if self.cloudcon.adapter.report:
            self.reportDir = self.cloudcon.adapter.report.logPath
        self.opened = self.cloudcon.opened
    
    def close(self):
        """cannot close the proxy"""
        pass
    
    def get_cloud_module(self):
        return self.cloudmodule
    
    def get_report_dir(self):
        return self.reportDir
    
    def job_add(self, params, logdata = None):
        return self.cloudcon.job_add(params, logdata)    
        
    def jobs_join(self, jids, timeout):
        """jobs join is not supported here as the server is not multithreaded"""
        return False                

    def jobs_map(self, params, mapargs, mapkwargs = None, logdata = None):
        return self.cloudcon.jobs_map(params, mapargs, mapkwargs, logdata)
    
    def jobs_result(self, jids, by_jid):
        return self.cloudcon.jobs_result(jids, by_jid)
    
    def jobs_kill(self, jids):
        return self.cloudcon.jobs_kill(jids)
    
    def jobs_delete(self, jids):
        return self.cloudcon.jobs_delete(jids)
    
    def jobs_info(self, jids, info_requested):
        return self.cloudcon.jobs_info(jids, info_requested)
    
    def is_simulated(self):
        return self.cloudcon.is_simulated()
    
    def needs_restart(self, **kwargs):
        return self.cloudcon.needs_restart(**kwargs)
    
    def send_request(self, url, post_values, get_values=None, logfunc=cloudLog.info):
        return self.cloudcon.send_request(url, post_values, get_values, logfunc)          
    
    def connection_info(self):
        return self.cloudcon.connection_info()
    
    def force_adapter_report(self):
        """
         Should the SerializationReport for the SerializationAdapter be coerced to be instantiated?
        """
        return self.cloudcon.force_adapter_report()
    
    def report_name(self):
        return self.cloudcon.report_name()       
    

def patch_cloud(cloudObjs, myPid):
    """Run on subprocesses to patch cloud with a list of cloudproxies
    Returns the primary cloud, constructed from the first cloudproxy"""
    from ..cloud import Cloud
    from .. import cloudinterface as ci
    
    report = None
    
    baseCl = None
    
    for cloudObj in cloudObjs:
        
        #construct cloud    
        sa = SerializingAdapter(cloudObj, isSlave = True)   
        reportDir = cloudObj.get_report_dir()
        if reportDir:
            sa.init_report()
            report = sa.report
            report.logPath = reportDir
            report.pid = myPid  #restore when supported 
            
        cl = Cloud(sa)
        
        modstr = cloudObj.get_cloud_module()
        __import__(modstr)
        mod = sys.modules[modstr]
            
        ci._bindcloud(mod, cl, 'proxy', immutable=True)
        if not baseCl:
            baseCl = cl
        
    return baseCl

def tail_file(filename, max_bytes=64000):
    """
    Returns the last max_bytes of a file. The storage is used to
    access the passed in filename.
    """
    
    size = os.path.getsize(filename)
    pos = max(0, size - max_bytes)
    
    f = file(filename)
    f.seek(pos)
    bytes = f.read()
    f.close()
    
    return bytes    
 
class MPConnection(CloudConnection, Pool):
    inKill = False
        
    #CONFIGURATION:
    c = """Number of subprocesses that cloud multiprocessing should utilize.
    Beware of using too low of a number of subprocesses, as deadlock can occur. 
    If set to 0, this will be set to the number of cores available on this system."""
    num_procs =  cc.mp_configurable('num_procs', default=8, comment=c)
    
    c = """Allow at most this factor*number subprocesses to be removed from the 
    internal priority queue and placed on the pipe."""
    io_buffer_factor =  cc.mp_configurable('io_buffer_factor', default=2.5, comment=c, hidden=True)
    
    c = """If set to true, job stdout and stderr will be redirected to files inside 
    the serialize_logging_path.  This option simulates PiCloud behavior and must be
    set to true for ``cloud.info`` to be able to request stdout or stderr information 
    about jobs.
    If set to false, job stdout and stderr will be the same as the parent process."""
    redirect_job_output = cc.mp_configurable('redirect_job_output', default=True, comment=c)
        
    c = """Maxmimum number of jobs that cloudmp should store.  Set to 0 for no limit.  
    This option only applies to cloud.mp and the cloud simulator.
    PiCloud recommends that any limit stays above 1024."""
    mp_cache_size = cc.mp_configurable('mp_cache_size_limit', default=0, comment=c)
            
    support_join = cc.mp_configurable('support_join', default = True, hidden=True,
                                     comment = 'Are high speed joins supported?')
    
    
        
    def __init__(self):
        self.openLock = threading.RLock()


    @property
    def pool(self):
        return self._pool
    
    def connection_info(self):
        dict = CloudConnection.connection_info(self)
        dict['connection_type'] = 'multiprocessing'        
        dict['num_procs'] = self.num_procs
        return dict    
    
    @property
    def cache(self):
        return self._cache
    
    def open(self):
        with self.openLock:
            
            if self._isopen:
                return            
            
            MPManager.register('ConnectionHook',MPConnectionHook) #force registration
            
            CloudConnection.open(self)        

            if self.mp_cache_size == 0:
                self._adapter.cloud.job_cache_size = 0
            else:
                self._adapter.cloud.job_cache_size = self.mp_cache_size
            self._adapter.cloud.result_cache_size = 0 #no limit in multiprocessing
                   
            #pool related
            self._setup_queues()
            
            if hasattr(Queue,'PriorityQueue'): #python 2.6+
                self._taskqueue = Queue.PriorityQueue()
            else:
                self._taskqueue = Queue.Queue()
            self._cache = {}
            self._state = pool.RUN
                    
            self.jobDeps = {}  #value is list of jobs that key depends on
            self.jobInvDeps = {} #value is list of jobs that depend on key
    
            #proxy objects
                    
            #Don't spawn more than one copy!
            if MPManager.myInstance:
            
                cloudmanager = MPManager.myInstance
            else:
                cloudmanager = MPManager()         
                cloudmanager.start_threaded()
            self.cloud_manager = cloudmanager
                
            #cloud-hooks. We must bind cloud and cloud.mp modules correctly
            self.cloudHooks = []        
            modnames = ['cloud', 'cloud.mp']
            
            for modname in modnames:
                try:
                    cloudmod =  sys.modules[modname]
                except KeyError:
                    cloudLog.error('Cannot find %s in sys.modules' % modname)
                else:
                    cloud = getattr(cloudmod,'__cloud')
                    adapter = cloud.adapter
                    conn = adapter.connection                                
                                        
                    if adapter.report:
                        reportDir = adapter.report.logPath
                    else:
                        reportDir = None
                    
                    chook = cloudmanager.ConnectionHook(conn, modname, reportDir)                    
                    
                    #first cloudhook is 'primary' -- i.e. this object
                    if conn == self:
                        self.cloudHooks.insert(0, chook)
                    else:
                        self.cloudHooks.append(chook)                
                    
            if not self.num_procs:
                try:
                    self.num_procs = cpu_count()
                except NotImplementedError:
                    self.num_procs = 1
    
            self._pool = []
            self.nextpid = 2
            
            for _ in xrange(self.num_procs):
                self._genprocess()        
    
            self._task_sem = threading.BoundedSemaphore(int(self.io_buffer_factor*self.num_procs))
    
            self._task_handler = threading.Thread(
                target=MPConnection._cloud_handle_tasks,
                name = '_cloud_handle_tasks',
                args=(self._taskqueue, self._task_sem, self._quick_put, self._outqueue, self._pool)
                )
            self._task_handler.daemon = True
            self._task_handler._state = pool.RUN
            
            self._task_handler.start()                
    
            self._result_handler = threading.Thread(
                target=Pool._handle_results,
                name = '_cloud_handle_results',
                args=(self._outqueue, self._quick_get, self._cache)
                )
            self._result_handler.daemon = True
            self._result_handler._state = pool.RUN
            self._result_handler.start()
            
            self._process_monitor_cv = threading.Condition()
            self._process_monitor = threading.Thread(
                target=self._monitor_pool,
                name = 'monitor_pool',
                args=()
                )            
            self._process_monitor.daemon = True
            self._process_monitor._state = pool.RUN
            self._process_monitor.start()
            
    
            self._terminate = Finalize(
                self, self._terminate_pool,
                args=(self._taskqueue, self._inqueue, self._outqueue, self._pool,
                      self._process_monitor, self._task_handler, self._result_handler, self._cache),
                exitpriority=15
                )
                
            self._worker_handler = self._process_monitor 

            self.dependency_lock = threading.RLock()
            self.job_kill_lock = threading.RLock()            
            
            #Joining  - A very simple implementation is used:
            #    jobs_join sleeps on join_cv until the jobs are not done
            #    Whenever a job completes, it wakes everyone sleeping on the join_cv 
            # This is not the most efficient system, but with the limited number of jobs 
            # being run, it doesn't matter
            if self.support_join:                
                self.join_cv = threading.Condition()
            else:
                self.join_cv = None
            
            #setupLogHandlers()
            cloudLog.info('Started Cloud multiprocessing with %d subprocesses' % self.num_procs)
            
            
   
    def _genprocess(self):
        from .. import cloudinterface
        w = self.Process(
            target=cloud_worker,
            args=(self._inqueue, self._outqueue, self.cloudHooks,  
                  self.nextpid, self.redirect_job_output,
                  self.adapter.serializeDebugging, self.adapter.serializeLogging,
                  cloudinterface.last_api_key, cloudinterface.last_api_secretkey)
            )
        self._pool.append(w)
        w.name = w.name.replace('Process', 'CloudPoolWorker')
        w.daemon = True
        w.start()
        w.cloudPid = self.nextpid #extra info
        self.nextpid +=1 

    def _kill_job(self, jid):        
        #TODO: We can abort early if we access the pqueue
                            
        job = self._cache.get(jid)
        if not job: #already done
            job = self.get_job_item(jid)  
            return job.status
        job.killMe = True #terminate job when it starts processing
        if job.status in ran_statuses:
            return job.status
        if job.status == status_processing: #important: this check must occur AFTER killMe set!
            self._do_job_kill(job)
        return status_killed #will be killed?               
        
    def _do_job_kill(self, job):
        """Kill process owned by job if necessary"""        
        
        #first we try a SIGINT which may raise an exception and get us out of here
        #not supported on windows when running python versions < 2.7
        if sys.platform != 'win32' or sys.version_info >= (2.7):            
            cloudLog.info('Sending SIGINT to job %d on pid %d. real pid' % (job.jid, job.pid))
            for proc in self._pool:
                if proc.cloudPid == job.pid:                      
                    os.kill(proc.ident, signal.SIGINT)
        
        time.sleep(1.0)
        cloudLog.info('Job is currently %s' % job.status)
        if job.status != status_processing: #sigint worked!
            return
            
        with self.job_kill_lock:            
            
            re_enter = False
            if self.inKill:
                re_enter = True
            elif self._outqueue._wlock:
                self._outqueue._wlock.acquire() #lock down write backs (block corruption) 
            #the write lock also holds up subprocesses on reading        
            killed = False                            
            self.inKill = True
            try:
                timeout = 0.1
                racquire = self._inqueue._rlock.acquire
                cloudLog.info('Initiating shutdown of job %d on pid %d' % (job.jid, job.pid))
                if not re_enter:
                    for _ in xrange(3): #a few tries
                        if racquire(block=True, timeout = timeout): #lock down lock job process reads
                            break
                        cloudLog.debug('Kill job %d on pid %d: Could not get lock - injecting backoff messages' % (job.jid, job.pid))                
                        for _ in xrange(self.num_procs*2):
                            self._quick_put(sleep_task_flag) #pump channel with backoff messages
                        timeout = 0.5 #highly likely to succeed now  
                cloudLog.debug('Kill job %d on pid %d: Acquired subprocess input pipe' % (job.jid, job.pid))         
                if job.status in ran_statuses: #verify again
                    return
                
                #Clear out result queue
                job.killMe = False #block re-entry on this job
                self._clear_results()            
                job.killMe = True    
                
                cloudLog.debug('Kill job %d on pid %d: Flushed result queue. jobs is %s' % (job.jid, job.pid, job.status))
                if job.status in ran_statuses: #verify one last time before killing..
                    return                            
                
                #At this point the result queue is clear AND no one can write anymore - kill the process            
                for proc in self._pool:
                    if proc.cloudPid == job.pid:
                        cloudLog.debug('terminating job %d on pid %d', job.jid, job.pid)
                        proc.terminate()
                        self._pool.remove(proc)
                        killed = True                    
                        break
            finally:      
                cloudLog.debug('termination done (killed=%s). releasing locks? %s', killed, not re_enter)          
                if not re_enter:
                    self.inKill = False
                    self._inqueue._rlock.release()
                    if self._outqueue._wlock:
                        self._outqueue._wlock.release()
                    cloudLog.debug('Locks are released!')
            if killed:
                cloudLog.debug('Job was terminated -- respawning, my state is %s. result handler state is %s', self._state, self._task_handler._state)
                job._aborted(status_killed)                
                if self._state or self._task_handler._state:
                    cloudLog.debug('not respawning due to non running state!')
                    return  #get out of here
                self._genprocess()
                cloudLog.debug('process respawned state is %s', self._state)

    def _clear_results(self):
        """Clear results from main queue and process
        Windows note: There appears to be an issue with calling poll() on named pipes in windows
        if another pipe is receiving.  Consequently, we must just sleep and hope that the taskmanager clears        
        """
        get = self._quick_get
        poll = self._outqueue._reader.poll        
        
        thread = threading.current_thread()
        
        if sys.platform != 'win32':
            cloudLog.debug('JobKiller: Clearing queue, entering loop? %s Am I result handler? %s' % (poll(),thread == self._result_handler))
        else: #can't call poll
            cloudLog.debug('JobKiller: Clearing queue,Am I result handler? %s' % (thread == self._result_handler))
        #print 'JobKiller: Clearing queue, entering loop? %s Am I result handler? %s' % (poll(),thread == self._result_handler)
        
        
        if thread != self._result_handler: 
            if sys.platform == 'win32': #we cannot safely call poll on windows
                time.sleep(2.0) #hopefully long enough
                return
            else:
                while poll():  #sleep until result handler gets all jobs
                    time.sleep(0.1)
            return
        
        #otherwise, I am the result handler -- clear the queue
        while poll():            
            try:
                #print 'access task'
                task = get()
                #print 'got task %s' % str(task)
            except (IOError, EOFError):
                debug('clearResults got EOFError/IOError -- exiting')
                return

            if thread._state:
                assert thread._state == pool.TERMINATE
                debug('clearResults on terminating')
                break

            if task is None:
                debug('clearResults got sentinel. reinjecting')
                self._outqueue._writer.send(None) #re-inject sentinel
                break

            job, i, obj = task
            try:
                #print 'cache set'
                self._cache[job]._set(i, obj)
                #print 'end set'
            except KeyError:
                pass   
            
    def _monitor_pool(self):
        """Monitor pool for unexpected terminations."""        
        while True:

            with self._process_monitor_cv:
                self._process_monitor_cv.wait(1.5)
            
            if not pool or self._state in (pool.CLOSE, pool.TERMINATE):
                    #pool may be none during system shutdown
                    break                            
            with self.job_kill_lock: #ensure that no job killing is occuring
                #double check once lock granted (abort if task_handler is terminating)
                if not pool or self._state in (pool.CLOSE, pool.TERMINATE) or self._task_handler._state:
                    #pool may be none during system shutdown
                    break            
                
                for p in self._pool[:]:
                    if not p.is_alive(): #sys.exit event!                        
                        pid = p.cloudPid
                        cloudLog.info('Killed process (%s) detected -- respawning! state is %s', pid, self._state)
                        for _, job in self._cache.items(): #find/delete relevant jobs
                            if job.status == status_processing and job.pid == pid:
                                job._aborted(status_error, CloudException('Job was killed with sys.exit(). Please do not use sys.exit() within cloud called functions'))
                        self._pool.remove(p)
                        self._genprocess() #restart process
    
    @staticmethod
    def _cloud_handle_tasks(taskqueue, task_sem, put, outqueue, pool):
        """Uses semaphore to allow pqueue to run"""
        thread = threading.current_thread()

        for _, taskseq in iter(taskqueue.get, None):
            for _, task in enumerate(taskseq):
                if thread._state:
                    debug('task handler found thread._state != RUN')
                    break
                try:
                    debug('task handler ack sem')
                    task_sem.acquire()
                    debug('task handler got sem')
                    #print 'put', task
                    put(task)
                    debug('task handler put done')
                    #print 'put done'
                except IOError:
                    debug('could not put task on queue')
                    break
            else:
                debug('task handler waiting...')
                continue
            break
        else:
            debug('task handler got sentinel')


        try:
            # tell result handler to finish when cache is empty
            debug('task handler sending sentinel to result handler')
            outqueue.put(None)

            # tell workers there is no more work
            debug('task handler sending sentinel to workers')
            for _ in pool:
                put(None)
        except IOError:
            debug('task handler got IOError when sending sentinels')

        debug('task handler exiting')
        
    @classmethod
    def _terminate_pool(cls, taskqueue, inqueue, outqueue, worker_pool,
                        worker_handler, task_handler, result_handler, cache):
        """This is taken from Python2.7.1 terminate_pool
        Added here as python2.7.2 changes some things
        """
        
        # this is guaranteed to only be called once
        debug('finalizing pool')

        task_handler._state = pool.TERMINATE
        worker_handler._state = pool.TERMINATE
        taskqueue.put(None)                 # sentinel

        debug('helping task handler/workers to finish')
        cls._help_stuff_finish(inqueue, task_handler, len(worker_pool))

        assert result_handler.is_alive() or len(cache) == 0

        result_handler._state = pool.TERMINATE
        outqueue.put(None)                  # sentinel

        if worker_pool and hasattr(worker_pool[0], 'terminate'):
            debug('terminating workers')
            for p in worker_pool:
                p.terminate()

        debug('joining task handler')
        task_handler.join(1e100)

        debug('joining result handler')
        result_handler.join(1e100)

        if worker_pool and hasattr(worker_pool[0], 'terminate'):
            debug('joining pool workers')
            for p in worker_pool:
                p.join()
        debug('everything shut down!')
        

    
    def notify_job_done(self, jid):
        job = self._cache[jid]
    
        #dispatch dependencies here                
        with self.dependency_lock:
            errored = job.status in error_statuses
            depends = self.jobInvDeps.get(job,[])
            for dep in depends:                
                indep = self.jobDeps[dep]
                del indep[indep.index(job)]
                if errored:
                    if dep.status not in ran_statuses:
                        dep._aborted(status_stalled)              
                if not indep: 
                    del self.jobDeps[dep]
                    if dep.status not in ran_statuses:
                        self._queue_job(dep)
                    
            if depends:
                del self.jobInvDeps[job]
                
        status = job.status
        exception = job.result if job.status != status_done else None
        
        loginfo = job.logname, job.logcnt, job.logpid
        
        if job.deleted:
            pass
        elif job.status == status_done:
            self.adapter.cloud.cacheManager.putCached(jid, status=status, exception = exception, 
                                                      runtime=job.runtime, loginfo = loginfo, result = job.result)
        else:
            self.adapter.cloud.cacheManager.putCached(jid, status=status, exception = exception, 
                                                      runtime=job.runtime, loginfo = loginfo)   
            
        if self.join_cv:
            with self.join_cv:
                self.join_cv.notifyAll()         
                                 
        
    def _queue_job(self, job):
        job._status = status_new
        self._taskqueue.put((job.priority, [(job.jid, job.func, job.args, job.kwds, 
                                             job.result_serialization_level, job.logname, job.logpid, job.logcnt)]))

    def cloud_apply_async(self, func, args=(), kwds={}, func_name = None, 
                          label = None, dependencies = [], priority = 5,
                          result_serialization_level = 0, 
                          logname="", logpid = 1, logcnt=0):
        '''
        Asynchronous equivalent of `apply()` builtin
        No callback support!!
        '''
        assert self._state == pool.RUN
        job = CloudJob(self, func, args, kwds, func_name, label, priority,
                       result_serialization_level, logname, logpid, logcnt)
        
        with self.dependency_lock:
            try:
                depJobs = [self.get_job_item(dep) for dep in maybe_xrange_iter(dependencies)]
            except CloudException ,e:
                e.parameter = 'This dependency does not exist.'                 
                raise e       
            runningDeps = []            
            for dep in depJobs:
                status =  dep.status                
                if status in error_statuses:
                    job._aborted(status_stalled)
                    break
                elif status not in ran_statuses:
                    runningDeps.append(dep)            
            if runningDeps:                 
                job._status = status_waiting
                self.jobDeps[job] = runningDeps  #job depends on runningDeps
                for indep in runningDeps:
                    depList = self.jobInvDeps.setdefault(indep, [])
                    depList.append(job)
        if job.status == status_new:
            self._queue_job(job)
        return job

    
    def close(self):
        CloudConnection.close(self)           
        cloudLog.debug('Terminating process pool')
        with self.job_kill_lock:  #avoid race condition with job killing
            self.terminate()
        cloudLog.debug('Notifying process monitor. state now %s', self._state)
        with self._process_monitor_cv:
            self._process_monitor_cv.notify() #force shutdown
        self.cloud_manager.close()
        self._process_monitor.join()    
  
    @staticmethod
    def _getparams(job):
        dependencies = job.get('depends_on',[]) 
        label = job.get('job_label',None)
        priority = job.get('job_priority',5) 
        func_name = job['func_name']
        result_serialization_level = job['fast_serialization']
        return func_name, dependencies, label, priority, result_serialization_level        
        
    
    def job_add(self, params, logdata = None):
        func = params['func']
        args = params['args']
        kwargs = params['kwargs']
        
        func_name, dependencies, label, priority, result_serialization_level = MPConnection._getparams(params)                                       
        
        logprefix, logpid, logcnt = logdata
        
        logname = logprefix + '.%sresult'        
        
        
        #print 'START running add... %s' % params
        #print 'add'
        thejob = self.cloud_apply_async(func=func, args=args, kwds=kwargs, func_name = func_name, 
                                        label = label, dependencies = dependencies, priority = priority,  
                                        result_serialization_level = result_serialization_level,
                                        logname=logname, logpid = logpid, logcnt = logcnt)
        #print 'done'
        #print 'DONE running add... %s' % params
        
        return thejob.jid
            
    def jobs_map(self, params, mapargs, mapkwargs = None, logdata = None):
        """Treat each element of argList as func(*element)
        """
        if not mapargs and not mapkwargs:
            raise ValueError('either mapargs or mapkwargs must be non-None')
        
        func = params['func']

        func_name, dependencies, label, priority, result_serialization_level = MPConnection._getparams(params)
        
        logprefix, logpid, logcnt = logdata
        
        ctr = 0
        outjids = []
        
        if not mapargs:
            mapargs = itertools.repeat(())
        if not mapkwargs:
            mapkwargs = itertools.repeat({})
        
        for args, kwargs in itertools.izip(mapargs, mapkwargs): 
            
            logname = ''.join([logprefix,'.%sjobresult_',str(ctr)])
            thejob = self.cloud_apply_async(func=func, args=args, kwds=kwargs, func_name = func_name, 
                                        label = label, dependencies = dependencies, priority = priority,
                                        result_serialization_level = result_serialization_level, 
                                        logname=logname, logpid = logpid, logcnt=logcnt)
            outjids.append(thejob.jid)
            ctr+=1
        return outjids
            
    def get_job_item(self, jid):
        """Retrieve multiprocessing job from either local or cloud cache"""            
        job = self._cache.get(jid)        
        if job:
            if job.deleted:
                raise CloudException('Does not exist or was purged from cache', jid=jid)
            return job
        #been deleted -- access cache
        job = self.adapter.cloud.cacheManager._getJob(jid)
        if job:
            return job
        raise CloudException('Does not exist or was purged from cache', jid=jid)
        
    def jobs_join(self, jids, timeout=None):
        """cv based joining
        Breaks when all jobs done or immediately if an error occurs or if timeout reached        
        """
        if not self.join_cv:
            return False        
        start_time = time.time()
        while True:
            statuses = []
            hit_error = False
            
            can_timeout = timeout is not None and time.time() - start_time > timeout
            
            for jid in maybe_xrange_iter(jids):
                job = self.get_job_item(jid)                
                if not hit_error and not can_timeout and job.status not in ran_statuses:
                    break                
                if job.status in error_statuses:
                    hit_error = True
                statuses.append(job.status)                
            else: #loop is done
                return statuses

            if can_timeout:
                return statuses
            
            with self.join_cv:
                if timeout is not None:
                    waitTime = max(0.1,timeout - (time.time() - start_time) + 0.1) 
                else:
                    waitTime = None
                self.join_cv.wait(waitTime)                        
    
    def jobs_result(self, jids, by_jid):
        def patchResult(jid):
            job = self.get_job_item(jid)
            
            result = job.result
            if not result:
                raise CloudException('Result was purged from cache', jid=jid)            
            return result
            
        return {'data': [patchResult(jid) for jid in maybe_xrange_iter(jids)],
                'interpretation': [{'datatype': 'python_pickle'} for jid in maybe_xrange_iter(jids)]}
    
    def jobs_kill(self, jids):
        target_jids = maybe_xrange_iter(jids) if jids else list(self._cache.keys())
        return [self._kill_job(jid) for jid in target_jids]        
    
    def jobs_delete(self, jids):
        #Kill anything still running
        for jid in  maybe_xrange_iter(jids):
            job = self.get_job_item(jid)
            if isinstance(job, CloudJob) and job.status not in ran_statuses:
                job.deleted = True
                self._kill_job(jid)
    
    def is_simulated(self):        
        return True 
    
    def needs_restart(self, **kwargs):        
        return False #maybe?    
    
    @staticmethod
    def patch_exception(job):
        """Helper for join/status"""
        if isinstance(job, CloudJob):
            return job.result if job.status == status_error else None
        else:  #from cache
            return job.exception    
    
    def jobs_info(self, jids, info_requested):
        jobs = [self.get_job_item(jid) for jid in maybe_xrange_iter(jids)]
        outdict = {}
        errMsg = 'redirect_job_output must be enabled in cloudconf.py to access stdout'
        for jid, job in itertools.izip(maybe_xrange_iter(jids),jobs):
            dct = {}
            if 'status' in info_requested:
                dct['status'] = job.status
                        
            if 'runtime' in info_requested:
                dct['runtime'] = job.runtime  #runtime is a property in CloudJob -- and therefore dynamically changing
                
            if 'exception' in info_requested:
                dct['exception'] = self.patch_exception(job)
            
            if 'stdout' in info_requested or 'stderr' in info_requested:
                report = self.adapter.report  
                if isinstance(job, CloudJob):
                    logname, logcnt, logpid = job.logname, job.logcnt, job.logpid                    
                else:
                    logname, logcnt, logpid = job.loginfo
                if not report:
                    raise CloudException(errMsg, jid=jid)
                if 'stdout' in info_requested:
                    stdoutFile = report.get_report_file(logname, '.stdout', logcnt, logpid)
                    try:                        
                        dct['stdout'] = tail_file(stdoutFile)
                    except (IOError, OSError):
                        if not isinstance(job, CloudJob): # cached - so job is done
                            raise CloudException('Cloud not load %s. %s' % (stdoutFile, errMsg), jid=jid)                        
                if 'stderr' in info_requested:
                    stderrFile = report.get_report_file(logname, '.stderr', logcnt, logpid)
                    try:                        
                        dct['stderr'] = tail_file(stderrFile)
                    except (IOError, OSError):
                        if not isinstance(job, CloudJob): # cached - so job is done
                            raise CloudException('Cloud not load %s. %s' % (stderrFile, errMsg), jid=jid)
                        
            # n/a results for simulation
            # todo: should be able to access profile
            # todo: consider saving created/endtime/logging
            non_sim = ['created', 'pilog', 'code_version', 'env', 'vol', 'finished', 'logging']
            for key in non_sim:
                if key in info_requested:
                    dct[key] = None 
                                    
            outdict[jid] = dct         
        return outdict                                   
    
    def force_adapter_report(self):
        """
        Should the SerializationReport for the SerializationAdapter be coerced to be instantiated?
        """
        return self.redirect_job_output
        
    def report_name(self):
        return 'Multiprocessing'    

